# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""E2E tests for the quantizer for model with add."""

import json

from absl.testing import parameterized
import absl.testing.absltest as absltest
import numpy as np
import os
import io

from ai_edge_litert.tools import flatbuffer_utils
from ai_edge_quantizer import quantizer
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils
from ai_edge_quantizer.utils import tfl_interpreter_utils
from ai_edge_litert import schema_py_generated as schema  # pylint:disable=g-direct-tensorflow-import

_RNG = np.random.default_rng(42)


# test_with_decomposition.mlir
# func @main(%arg0, %arg1) ->  {
#   %0 = composite "npu_call" %arg0, %arg1 {decomposition = @decomp}
#   return %0
# }
# func private @decomp(%arg0: , %arg1: ) ->  {
#   %0 = tfl.add %arg0, %arg1
#   return %0
# }

# test_with_kernel.mlir (subset)
# %12 = stablehlo.composite "scaled_dot_product_attention" ...


def _get_calibration_data(
    model: schema.ModelT,
    name_pref: str,
    subgraph_index: int = 0,
    num_samples: int = 2,
):
  sg: schema.SubGraphT = model.subgraphs[subgraph_index]
  inputs = [sg.tensors[i] for i in sg.inputs]

  data = []

  for _ in range(num_samples):
    samp_data = {}
    for i, inp in enumerate(inputs):
      inp: schema.TensorT = inp
      name = f'{name_pref}{i}'
      samp_data[name] = _RNG.uniform(size=inp.shape).astype(np.float32)

    data.append(samp_data)

  return {
      tfl_interpreter_utils.DEFAULT_SIGNATURE_KEY: data,
  }


class CompositeTest(parameterized.TestCase):

  @property
  def output_tolerance(self) -> float:
    return 1e-4

  def assertNotQuantiezedType(self, tensor: schema.TensorT):
    self.assertIn(
        tensor.type, [schema.TensorType.FLOAT32, schema.TensorType.INT32]
    )

  @parameterized.parameters(
      '../../recipes/default_a8w8_recipe.json',
      '../../recipes/default_a16w8_recipe.json',
  )
  def test_composite_with_decomposition(self, recipe_path):
    model = test_utils.get_path_to_datafile('../models/simple_composite.tflite')
    qt = quantizer.Quantizer(model)

    recipe_path = test_utils.get_path_to_datafile(recipe_path)
    with open(recipe_path, 'r') as f:
      recipe_json = json.load(f)
    qt.load_quantization_recipe(recipe_json)

    f_model = tfl_flatbuffer_utils.read_model(model)
    self.assertTrue(qt.need_calibration)
    calib_data = _get_calibration_data(f_model, 'arg')
    calibration_result = qt.calibrate(calib_data)

    result = qt.quantize(calibration_result)
    q_model = tfl_flatbuffer_utils.read_model(result.quantized_model)
    for sg in q_model.subgraphs:
      for tensor in sg.tensors:
        self.assertIsNotNone(tensor.quantization)

    test_data = _get_calibration_data(f_model, 'arg')
    comparison = qt.validate(error_metrics='mse', test_data=test_data)
    comparison = comparison.get_all_tensor_results()
    output_mse = comparison['output']
    self.assertLess(output_mse, self.output_tolerance)

  @parameterized.parameters(
      '../../recipes/default_a8w8_recipe.json',
      '../../recipes/default_a16w8_recipe.json',
  )
  def test_composite_with_kernel(self, recipe_path):
    model = test_utils.get_path_to_datafile('../models/sdpa_composite.tflite')
    qt = quantizer.Quantizer(model)

    recipe_path = test_utils.get_path_to_datafile(recipe_path)
    with open(recipe_path, 'r') as f:
      recipe_json = json.load(f)
    qt.load_quantization_recipe(recipe_json)
    self.assertTrue(qt.need_calibration)

    f_model = tfl_flatbuffer_utils.read_model(model)
    calib_data = _get_calibration_data(f_model, 'args_')
    calibration_result = qt.calibrate(calib_data)

    result = qt.quantize(calibration_result)
    q_model = tfl_flatbuffer_utils.read_model(result.quantized_model)

    main_sg = q_model.subgraphs[0]
    for op in main_sg.operators:
      if 'COMPOSITE' in flatbuffer_utils.opcode_to_name(
          q_model, op.opcodeIndex
      ):
        for inp in op.inputs:
          tensor = main_sg.tensors[inp]
          self.assertNotQuantiezedType(tensor)
        for out in op.outputs:
          tensor = main_sg.tensors[out]
          self.assertNotQuantiezedType(tensor)

    sdpa_sg = q_model.subgraphs[1]
    for tensor in sdpa_sg.tensors:
      self.assertNotQuantiezedType(tensor)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""E2E tests for the quantizer for model with transpose."""

from typing import Any

from absl.testing import parameterized
import absl.testing.absltest as absltest
import numpy as np

from ai_edge_quantizer import qtyping
from ai_edge_quantizer import quantizer
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils
from ai_edge_quantizer.utils import tfl_interpreter_utils

_OpExecutionMode = qtyping.OpExecutionMode
_OpName = qtyping.TFLOperationName
_TensorQuantConfig = qtyping.TensorQuantizationConfig
_OpQuantConfig = qtyping.OpQuantizationConfig

_RNG = np.random.default_rng(66)


def _get_dummy_data(
    num_samples: int, dtype: np.dtype = np.float32
) -> list[dict[str, Any]]:
  data = []
  for _ in range(num_samples):
    data.append({'input_2': _RNG.uniform(size=(1, 2, 3, 4)).astype(dtype)})
  return data


def _get_calibration_data(
    num_samples: int = 128, dtype: np.dtype = np.float32
) -> list[dict[str, Any]]:
  calibration_samples = _get_dummy_data(num_samples, dtype)
  calibration_data = {
      tfl_interpreter_utils.DEFAULT_SIGNATURE_KEY: calibration_samples,
  }
  return calibration_data


def _get_test_data(
    num_samples: int = 8, dtype: np.dtype = np.float32
) -> list[dict[str, Any]]:
  return _get_calibration_data(num_samples, dtype)


class FloatTransposeTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.float_model_path = test_utils.get_path_to_datafile(
        '../models/single_transpose.tflite'
    )
    self._quantizer = quantizer.Quantizer(self.float_model_path)

  @parameterized.named_parameters(
      dict(
          testcase_name='a8w8',
          recipe_path='../../recipes/default_a8w8_recipe.json',
          input_output_dtype_code=9,  # int8.
      ),
      dict(
          testcase_name='a16w8',
          recipe_path='../../recipes/default_a16w8_recipe.json',
          input_output_dtype_code=7,  # int16.
      ),
  )
  def test_transpose_model_full_integer(
      self, recipe_path, input_output_dtype_code
  ):
    recipe_path = test_utils.get_path_to_datafile(recipe_path)
    self._quantizer.load_quantization_recipe(recipe_path)
    self.assertTrue(self._quantizer.need_calibration)
    calibration_result = self._quantizer.calibrate(_get_calibration_data())
    quantization_result = self._quantizer.quantize(calibration_result)

    # Check tensor dtypes.
    quantized_model = tfl_flatbuffer_utils.read_model(
        quantization_result.quantized_model
    )
    self.assertLen(quantized_model.subgraphs, 1)
    subgraph = quantized_model.subgraphs[0]
    subgraph_tensors = subgraph.tensors
    input_tensor = subgraph_tensors[subgraph.inputs[0]]
    output_tensor = subgraph_tensors[subgraph.outputs[0]]
    # Check input/output tensor type.
    # See schema_py_generated.py for type code.
    self.assertEqual(input_tensor.type, input_output_dtype_code)
    self.assertEqual(output_tensor.type, input_output_dtype_code)

    comparison_result = self._quantizer.validate(
        error_metrics='mse', test_data=_get_test_data()
    )
    self._check_comparison_result(comparison_result, output_tolerance=1e-4)

  def _check_comparison_result(self, comparison_result, output_tolerance):
    # TODO: b/357959309 - Use comparison result directly for testing.
    comparison_result = comparison_result.get_all_tensor_results()
    output_mse = comparison_result['PartitionedCall:0']
    self.assertLess(output_mse, output_tolerance)


class IntegerTransposeTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.model_path = test_utils.get_path_to_datafile(
        '../models/single_transpose_int32.tflite'
    )
    self._quantizer = quantizer.Quantizer(self.model_path)

  @parameterized.parameters(
      '../../recipes/default_a8w8_recipe.json',
      '../../recipes/default_a16w8_recipe.json',
  )
  def test_quantize_integer_transpose(self, recipe_path):
    recipe_path = test_utils.get_path_to_datafile(recipe_path)
    self._quantizer.load_quantization_recipe(recipe_path)
    self.assertTrue(self._quantizer.need_calibration)
    calibration_result = self._quantizer.calibrate(
        _get_calibration_data(dtype=np.int32)
    )
    quantization_result = self._quantizer.quantize(calibration_result)
    quantized_model = tfl_flatbuffer_utils.read_model(
        quantization_result.quantized_model
    )
    self.assertLen(quantized_model.subgraphs, 1)
    subgraph = quantized_model.subgraphs[0]
    subgraph_tensors = subgraph.tensors
    self.assertLen(subgraph.inputs, 1)
    input_tensor = subgraph_tensors[subgraph.inputs[0]]
    output_tensor = subgraph_tensors[subgraph.outputs[0]]
    # Check input/output tensor type.
    # See schema_py_generated.py for type code.
    self.assertEqual(input_tensor.type, 2)  # int32.
    self.assertEqual(output_tensor.type, 2)  # int32.


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""quantize a given tensor."""

from typing import Optional, cast
import ml_dtypes
import numpy as np
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.transformations import transformation_utils
from ai_edge_litert import schema_py_generated  # pylint: disable=g-direct-tensorflow-import


# TODO: b/335014051 - Support distinguishing INT, FLOAT & UINT, BFLOAT.
def quant_params_to_tflite_type(
    bitwidth: int,
) -> Optional[schema_py_generated.TensorType]:
  """Given specifications from quant param return the corresponding TFLite dtype.

  Args:
    bitwidth: Bit width from UniformQuantParams.

  Returns:
    The corresponding TFLite tensor type.
  """
  if bitwidth == 4:
    return schema_py_generated.TensorType.INT4
  elif bitwidth <= 8:
    return schema_py_generated.TensorType.INT8
  elif bitwidth <= 16:
    return schema_py_generated.TensorType.INT16
  elif bitwidth <= 32:
    return schema_py_generated.TensorType.INT32
  elif bitwidth <= 64:
    return schema_py_generated.TensorType.INT64
  else:
    raise ValueError(f"Unsupported quant params: {bitwidth}")


def nonlinear_quant_params_to_tflite_type(
    bitwidth: int,
) -> Optional[schema_py_generated.TensorType]:
  """Given specifications from quant param return the corresponding tflite dtype.

  Args:
    bitwidth: bitwidth from NonLinearQuantParams

  Returns:
    the corresponding tflite tensortype
  """
  if bitwidth == 16:
    return schema_py_generated.TensorType.FLOAT16
  elif bitwidth == 32:
    return schema_py_generated.TensorType.FLOAT32
  else:
    raise ValueError(f"Unsupported nonlinear params: {bitwidth}")


def _perform_channelwise_quantization(
    transformation_input: transformation_utils.TransformationInput,
) -> schema_py_generated.QuantizationParametersT():
  """Perform channelwise quantization and fill the quantization parameters.

  Args:
    transformation_input: Input structure that contains all information needed
      for the transformation.

  Returns:
    The quantization parameters.
  """
  assert isinstance(
      transformation_input.quant_params, qtyping.UniformQuantParams
  )
  flatbuffer_quantization = schema_py_generated.QuantizationParametersT()
  flatbuffer_quantization.scale = list(
      transformation_input.quant_params.scale.flatten().astype(np.float32)
  )  # Flatbuffer requires scale as list[float].
  if transformation_input.quant_params.zero_point is not None:
    flatbuffer_quantization.zeroPoint = list(
        transformation_input.quant_params.zero_point.flatten().astype(np.int64)
    )  # Flatbuffer requires zeroPoint as list[int64]
  if transformation_input.quant_params.quantized_dimension is not None:
    flatbuffer_quantization.quantizedDimension = (
        transformation_input.quant_params.quantized_dimension
    )

  return flatbuffer_quantization


def _perform_blockwise_quantization(
    transformation_input: transformation_utils.TransformationInput,
) -> schema_py_generated.QuantizationParametersT():
  """Perform blockwise quantization and fill the quantization parameters.

  Args:
    transformation_input: Input structure that contains all information needed
      for the transformation.

  Returns:
    The quantization parameters.
  """
  assert isinstance(
      transformation_input.quant_params, qtyping.UniformQuantParams
  )
  flatbuffer_quantization = schema_py_generated.QuantizationParametersT()
  flatbuffer_quantization.detailsType = (
      schema_py_generated.QuantizationDetails.BlockwiseQuantization
  )
  tensor = transformation_input.subgraph.tensors[transformation_input.tensor_id]
  blockwise_details = schema_py_generated.BlockwiseQuantizationT()
  # Downcast and round the scale to fp16 with 7 bit mantissa.
  scale_tensor_id = transformation_utils.add_new_constant_tensor(
      tensor.name + b"_scales",
      transformation_input.quant_params.scale.astype(ml_dtypes.bfloat16).astype(
          np.float16
      ),
      schema_py_generated.TensorType.FLOAT16,
      transformation_input.subgraph,
      transformation_input.buffers,
  )
  blockwise_details.scales = scale_tensor_id
  # Blockwise quantization does not support zero point yet, so this points to
  # a -1 buffer index.
  # TODO: b/404909258 - Add optional zero point to blockwise quantization.
  blockwise_details.zeroPoints = -1
  blockwise_details.blockSize = transformation_input.quant_params.block_size
  flatbuffer_quantization.details = blockwise_details
  # TODO: b/443830202 - Hardcoding to 0 for now.
  flatbuffer_quantization.quantizedDimension = 0
  return flatbuffer_quantization


def quantize_tensor(
    transformation_input: transformation_utils.TransformationInput,
) -> qtyping.TransformationInfo:
  """Quantize the tensor at the tensor_id in the given subgraph.

  Args:
    transformation_input: Input structure that contains all information needed
      for the transformation.

  Returns:
    TransformationInfo:
      op_id: The producer index for tensor.
      num_ops_added: The total number of ops inserted by this operation, which
        is 0.
  """
  tensor = transformation_input.subgraph.tensors[transformation_input.tensor_id]
  # TODO: b/336385820 - Suppport quantize buffer directly when quantized_data
  # is not provided.
  if tensor.buffer:
    if transformation_input.quant_params.quantized_data is not None:
      transformation_input.buffers[tensor.buffer].data = (
          transformation_utils.pack_data(
              transformation_input.quant_params.num_bits,
              np.frombuffer(
                  cast(
                      np.ndarray,
                      transformation_input.quant_params.quantized_data,
                  ).tobytes(),
                  dtype=np.uint8,
              ).flatten(),
          )
      )

  if isinstance(transformation_input.quant_params, qtyping.UniformQuantParams):
    if transformation_input.quant_params.block_size == 0:
      flatbuffer_quantization = _perform_channelwise_quantization(
          transformation_input
      )
    else:
      flatbuffer_quantization = _perform_blockwise_quantization(
          transformation_input
      )
    tensor.quantization = flatbuffer_quantization
    tensor.type = quant_params_to_tflite_type(
        transformation_input.quant_params.num_bits
    )
  if isinstance(
      transformation_input.quant_params, qtyping.NonLinearQuantParams
  ):
    tensor.type = nonlinear_quant_params_to_tflite_type(
        transformation_input.quant_params.num_bits
    )

  return qtyping.TransformationInfo(
      0, num_ops_added=0, output_tensor_id=transformation_input.tensor_id
  )



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Utils for tests."""

import inspect as _inspect
import pathlib as _pathlib
import sys as _sys
from typing import Optional, Union

from absl.testing import parameterized

from ai_edge_quantizer import model_validator
from ai_edge_quantizer import qtyping
from ai_edge_quantizer import quantizer
from ai_edge_quantizer.utils import tfl_interpreter_utils

_ComputePrecision = qtyping.ComputePrecision
_OpName = qtyping.TFLOperationName
_TensorQuantConfig = qtyping.TensorQuantizationConfig
_OpQuantConfig = qtyping.OpQuantizationConfig
_AlgorithmName = quantizer.AlgorithmName
_Numeric = Union[int, float]


DEFAULT_ACTIVATION_QUANT_SETTING = _TensorQuantConfig(
    num_bits=8,
    symmetric=False,
    granularity=qtyping.QuantGranularity.TENSORWISE,
)
DEFAULT_WEIGHT_QUANT_SETTING = _TensorQuantConfig(
    num_bits=8,
    symmetric=True,
    granularity=qtyping.QuantGranularity.CHANNELWISE,
)


def get_static_activation_quant_setting(
    num_bits: int, symmetric: bool
) -> _TensorQuantConfig:
  return _TensorQuantConfig(
      num_bits=num_bits,
      symmetric=symmetric,
      granularity=qtyping.QuantGranularity.TENSORWISE,
  )


def get_static_op_quant_config(
    activation_config: _TensorQuantConfig = DEFAULT_ACTIVATION_QUANT_SETTING,
    weight_config: _TensorQuantConfig = DEFAULT_WEIGHT_QUANT_SETTING,
) -> _OpQuantConfig:
  return qtyping.OpQuantizationConfig(
      activation_tensor_config=activation_config,
      weight_tensor_config=weight_config,
      compute_precision=_ComputePrecision.INTEGER,
  )


def get_path_to_datafile(path):
  """Get the path to the specified file in the data dependencies.

  The path is relative to the file calling the function.

  Args:
    path: a string resource path relative to the calling file.

  Returns:
    The path to the specified file present in the data attribute of py_test
    or py_binary.

  Raises:
    IOError: If the path is not found, or the resource can't be opened.
  """
  data_files_path = _pathlib.Path(_inspect.getfile(_sys._getframe(1))).parent  # pylint: disable=protected-access
  return str((data_files_path / path).resolve())


class BaseOpTestCase(parameterized.TestCase):
  """Base class for op-level tests."""

  def quantize_and_validate(
      self,
      model_path: str,
      algorithm_key: _AlgorithmName,
      op_name: _OpName,
      op_config: _OpQuantConfig,
      num_validation_samples: int = 4,
      num_calibration_samples: Optional[int] = None,
      error_metric: str = 'mse',
      min_max_range: Optional[tuple[_Numeric, _Numeric]] = None,
  ) -> model_validator.ComparisonResult:
    """Quantizes and validates the given model with the given configurations.

    Args:
      model_path: The path to the model to be quantized.
      algorithm_key: The algorithm to be used for quantization.
      op_name: The name of the operation to be quantized.
      op_config: The configuration for the operation to be quantized.
      num_validation_samples: The number of samples to use for validation.
      num_calibration_samples: The number of samples to use for calibration. If
        None then it will be set to num_validation_samples * 8.
      error_metric: The error error_metric to use for validation.
      min_max_range: The min and max of the input range.

    Returns:
      The comparison result of the validation.
    """
    quantizer_instance = quantizer.Quantizer(model_path)
    quantizer_instance.update_quantization_recipe(
        algorithm_key=algorithm_key,
        regex='.*',
        operation_name=op_name,
        op_config=op_config,
    )
    if quantizer_instance.need_calibration:
      if num_calibration_samples is None:
        num_calibration_samples = num_validation_samples * 8
      calibration_data = tfl_interpreter_utils.create_random_normal_input_data(
          quantizer_instance._float_model_buffer,  # pylint: disable=protected-access
          num_samples=num_calibration_samples,
          min_max_range=min_max_range,
      )
      calibration_result = quantizer_instance.calibrate(calibration_data)
      quantization_result = quantizer_instance.quantize(calibration_result)
    else:
      quantization_result = quantizer_instance.quantize()
    test_data = tfl_interpreter_utils.create_random_normal_input_data(
        quantization_result.quantized_model,
        num_samples=num_validation_samples,
        min_max_range=min_max_range,
    )
    return quantizer_instance.validate(test_data, error_metric)

  def assert_model_size_reduction_above_min_pct(
      self,
      validation_result: model_validator.ComparisonResult,
      min_pct: float,
  ):
    """Checks the model size reduction (percentage) against the given expectation."""
    _, reduction_pct = validation_result.get_model_size_reduction()
    self.assertGreater(reduction_pct, min_pct)

  def assert_weights_errors_below_tolerance(
      self,
      validation_result: model_validator.ComparisonResult,
      weight_tolerance: float,
  ):
    """Checks the weight tensors' numerical behavior against the given tolerance."""
    self.assertNotEmpty(validation_result.available_signature_keys())
    for signature_key in validation_result.available_signature_keys():
      signature_result = validation_result.get_signature_comparison_result(
          signature_key
      )
      for result in signature_result.constant_tensors.values():
        self.assertLess(result, weight_tolerance)

  def assert_output_errors_below_tolerance(
      self,
      validation_result: model_validator.ComparisonResult,
      output_tolerance: float,
  ):
    """Checks the output tensor numerical behavior against the given tolerance."""
    self.assertNotEmpty(validation_result.available_signature_keys())
    for signature_key in validation_result.available_signature_keys():
      signature_result = validation_result.get_signature_comparison_result(
          signature_key
      )
      for result in signature_result.output_tensors.values():
        self.assertLess(result, output_tolerance)

  def assert_quantization_accuracy_and_size(
      self,
      algorithm_key: _AlgorithmName,
      model_path: str,
      op_name: _OpName,
      op_config: _OpQuantConfig,
      expected_model_size_reduction: float,
      weight_tolerance: float = 1e-4,
      output_tolerance: float = 1e-4,
      min_max_range: Optional[tuple[_Numeric, _Numeric]] = None,
  ):
    """Check if the quantization is successful and the result is valid."""
    validation_result = self.quantize_and_validate(
        model_path=model_path,
        algorithm_key=algorithm_key,
        op_name=op_name,
        op_config=op_config,
        min_max_range=min_max_range,
    )
    with self.subTest(name='ModelSizeReduction'):
      self.assert_model_size_reduction_above_min_pct(
          validation_result, expected_model_size_reduction
      )
    with self.subTest(name='WeightsErrors'):
      self.assert_weights_errors_below_tolerance(
          validation_result, weight_tolerance
      )
    with self.subTest(name='OutputErrors'):
      self.assert_output_errors_below_tolerance(
          validation_result, output_tolerance
      )

  def assert_quantization_accuracy(
      self,
      algorithm_key: _AlgorithmName,
      model_path: str,
      op_name: _OpName,
      op_config: _OpQuantConfig,
      num_validation_samples: int = 4,
      num_calibration_samples: Optional[int] = None,
      output_tolerance: float = 1e-4,
      min_max_range: Optional[tuple[_Numeric, _Numeric]] = None,
  ):
    """Checks if the output errors after quantization are within the tolerance."""
    validation_result = self.quantize_and_validate(
        model_path=model_path,
        algorithm_key=algorithm_key,
        num_validation_samples=num_validation_samples,
        num_calibration_samples=num_calibration_samples,
        op_name=op_name,
        op_config=op_config,
        min_max_range=min_max_range,
    )
    self.assert_output_errors_below_tolerance(
        validation_result, output_tolerance
    )



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Test insertion of the Decomposed Hadamard rotation ops."""

import pathlib
import numpy as np
import absl.testing.absltest as absltest
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.transformations import insert_decomposed_hadamard_rotation
from ai_edge_quantizer.transformations import transformation_utils
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils
from ai_edge_litert import schema_py_generated  # pylint: disable=g-direct-tensorflow-import

_TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile('..')


class InsertDecomposedHadamardRotationFullyConnectedTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    model_path = str(
        pathlib.Path(_TEST_DATA_PREFIX_PATH) / 'tests/models/single_fc_bias.tflite'
    )
    self.model = tfl_flatbuffer_utils.read_model(model_path)
    self.params = qtyping.UniformQuantParams(
        num_bits=8,
        quantized_dimension=None,
        scale=np.ones(1),
        zero_point=np.zeros(1),
        hadamard=qtyping.UniformQuantParams.HadamardRotationParams(
            random_binary_vector=np.ones(1),
            hadamard_size=2,
        ),
    )

  def test_raise_unsupported_qparams(self):
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: 'uniform quantization' in str(err)
    ):
      insert_decomposed_hadamard_rotation.insert_decomposed_hadamard_rotation(
          transformation_utils.TransformationInput(
              tensor_id=0,
              op_codes=self.model.operatorCodes,
              buffers=self.model.buffers,
              subgraph=self.model.subgraphs[0],
              producer=-1,
              consumers=[-1],
              quant_params=qtyping.NonLinearQuantParams(
                  num_bits=16, quantized_data=None
              ),
          )
      )

  def test_raise_missing_hadamard_data(self):
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: 'quantization params are not set' in str(err)
    ):
      insert_decomposed_hadamard_rotation.insert_decomposed_hadamard_rotation(
          transformation_utils.TransformationInput(
              tensor_id=0,
              op_codes=self.model.operatorCodes,
              buffers=self.model.buffers,
              subgraph=self.model.subgraphs[0],
              producer=-1,
              consumers=[-1],
              quant_params=qtyping.UniformQuantParams(
                  num_bits=8,
                  quantized_dimension=None,
                  scale=np.ones(1),
                  zero_point=np.zeros(1),
              ),
          )
      )

  def test_raise_non_float32_tensor(self):
    self.model.subgraphs[0].tensors[
        0
    ].type = schema_py_generated.TensorType.INT32
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: 'float32 tensors' in str(err)
    ):
      insert_decomposed_hadamard_rotation.insert_decomposed_hadamard_rotation(
          transformation_utils.TransformationInput(
              tensor_id=0,
              op_codes=self.model.operatorCodes,
              buffers=self.model.buffers,
              subgraph=self.model.subgraphs[0],
              producer=-1,
              consumers=[-1],
              quant_params=self.params,
          ),
      )

  def test_insert_decomposed_ops(self):
    # Insert Decomposed Hadamard ops before fully_connected
    info = (
        insert_decomposed_hadamard_rotation.insert_decomposed_hadamard_rotation(
            transformation_utils.TransformationInput(
                tensor_id=0,
                op_codes=self.model.operatorCodes,
                buffers=self.model.buffers,
                subgraph=self.model.subgraphs[0],
                producer=-1,
                consumers=[0],  # Consumer is the FC op
                quant_params=self.params,
            )
        )
    )
    subgraph = self.model.subgraphs[0]
    self.assertEqual(info.op_id, 0)
    self.assertEqual(info.num_ops_added, 3)
    # Model had 4 tensors, added 6 tensors (3 activations 3 constants).
    self.assertEqual(info.output_tensor_id, 9)
    self.assertLen(subgraph.tensors, 10)
    # Model had 1 op code, added RESHAPE and FC.
    self.assertLen(self.model.operatorCodes, 3)
    self.assertEqual(
        self.model.operatorCodes[1].builtinCode,
        schema_py_generated.BuiltinOperator.RESHAPE,
    )
    self.assertEqual(
        self.model.operatorCodes[2].builtinCode,
        schema_py_generated.BuiltinOperator.FULLY_CONNECTED,
    )

    # Op 0: RESHAPE
    reshape_op = subgraph.operators[0]
    self.assertEqual(
        self.model.operatorCodes[reshape_op.opcodeIndex].builtinCode,
        schema_py_generated.BuiltinOperator.RESHAPE,
    )
    self.assertEqual(reshape_op.inputs[0], 0)  # Graph input
    self.assertEqual(reshape_op.outputs[0], 5)  # Reshape output

    # Op 1: FULLY_CONNECTED
    fc_op = subgraph.operators[1]
    self.assertEqual(
        self.model.operatorCodes[fc_op.opcodeIndex].builtinCode,
        schema_py_generated.BuiltinOperator.FULLY_CONNECTED,
    )
    self.assertEqual(fc_op.inputs[0], 5)  # Reshape output
    self.assertEqual(fc_op.inputs[1], 6)  # Hadamard matrix tensor
    self.assertEqual(fc_op.outputs[0], 7)  # FC output

    # Op 2: RESHAPE (post)
    post_reshape_op = subgraph.operators[2]
    self.assertEqual(
        self.model.operatorCodes[post_reshape_op.opcodeIndex].builtinCode,
        schema_py_generated.BuiltinOperator.RESHAPE,
    )
    self.assertEqual(post_reshape_op.inputs[0], 7)  # FC output
    self.assertEqual(post_reshape_op.outputs[0], 9)  # Post Reshape output

    # Op 3: Original FULLY_CONNECTED
    orig_fc_op = subgraph.operators[3]
    self.assertEqual(
        self.model.operatorCodes[orig_fc_op.opcodeIndex].builtinCode,
        schema_py_generated.BuiltinOperator.FULLY_CONNECTED,
    )
    # Input to the original FC is the post reshape output
    self.assertEqual(orig_fc_op.inputs[0], 9)


class InsertDecomposedHadamardRotationEmbeddingLookupTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    model_path = str(
        pathlib.Path(_TEST_DATA_PREFIX_PATH) / 'tests/models/embedding_lookup.tflite'
    )
    self.model = tfl_flatbuffer_utils.read_model(model_path)
    self.params = qtyping.UniformQuantParams(
        num_bits=8,
        quantized_dimension=None,
        scale=np.ones(1),
        zero_point=np.zeros(1),
        hadamard=qtyping.UniformQuantParams.HadamardRotationParams(
            random_binary_vector=np.ones(1),
            hadamard_size=2,
        ),
    )

  def test_insert_decomposed_ops(self):
    # Insert Decomposed Hadamard ops after embedding_lookup
    info = (
        insert_decomposed_hadamard_rotation.insert_decomposed_hadamard_rotation(
            transformation_utils.TransformationInput(
                tensor_id=2,  # Output of embedding_lookup
                op_codes=self.model.operatorCodes,
                buffers=self.model.buffers,
                subgraph=self.model.subgraphs[0],
                producer=0,
                consumers=[-1],  # Output is a graph output
                quant_params=self.params,
            )
        )
    )
    subgraph = self.model.subgraphs[0]
    self.assertEqual(info.op_id, 1)
    self.assertEqual(info.num_ops_added, 3)
    # Model had 3 tensors, added 6 (3 activations 3 constants).
    self.assertEqual(info.output_tensor_id, 8)
    self.assertLen(subgraph.tensors, 9)
    # Model had 1 op code, added RESHAPE and FC.
    self.assertLen(self.model.operatorCodes, 3)

    # Op 0: EMBEDDING_LOOKUP (Original)
    # Op 1: RESHAPE
    reshape_op = subgraph.operators[1]
    self.assertEqual(reshape_op.inputs[0], 2)  # Embedding lookup output
    self.assertEqual(reshape_op.outputs[0], 4)

    # Op 2: FULLY_CONNECTED
    fc_op = subgraph.operators[2]
    self.assertEqual(fc_op.inputs[0], 4)
    self.assertEqual(fc_op.inputs[1], 5)  # Hadamard matrix
    self.assertEqual(fc_op.outputs[0], 6)

    # Op 3: RESHAPE (post)
    post_reshape_op = subgraph.operators[3]
    self.assertEqual(post_reshape_op.inputs[0], 6)
    self.assertEqual(post_reshape_op.outputs[0], 8)

    # Check graph output
    self.assertIn(8, subgraph.outputs)
    self.assertNotIn(2, subgraph.outputs)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""E2E tests for the quantizer for model with transpose."""

from typing import Any

from absl.testing import parameterized
import absl.testing.absltest as absltest
import numpy as np

from ai_edge_quantizer import qtyping
from ai_edge_quantizer import quantizer
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils
from ai_edge_quantizer.utils import tfl_interpreter_utils

_OpExecutionMode = qtyping.OpExecutionMode
_OpName = qtyping.TFLOperationName
_TensorQuantConfig = qtyping.TensorQuantizationConfig
_OpQuantConfig = qtyping.OpQuantizationConfig

_RNG = np.random.default_rng(66)


def _get_dummy_data(
    num_samples: int, dtype: np.dtype = np.float32
) -> list[dict[str, Any]]:
  data = []
  for _ in range(num_samples):
    data.append({'input_1': _RNG.uniform(size=(2, 3)).astype(dtype)})
  return data


def _get_calibration_data(
    num_samples: int = 128, dtype: np.dtype = np.float32
) -> list[dict[str, Any]]:
  calibration_samples = _get_dummy_data(num_samples, dtype)
  calibration_data = {
      tfl_interpreter_utils.DEFAULT_SIGNATURE_KEY: calibration_samples,
  }
  return calibration_data


def _get_test_data(
    num_samples: int = 8, dtype: np.dtype = np.float32
) -> list[dict[str, Any]]:
  return _get_calibration_data(num_samples, dtype)


class SumTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.float_model_path = test_utils.get_path_to_datafile(
        '../models/single_sum.tflite'
    )
    self._quantizer = quantizer.Quantizer(self.float_model_path)

  @parameterized.named_parameters(
      dict(
          testcase_name='int8_quantized',
          recipe_path='../../recipes/default_a8w8_recipe.json',
          tensor_type=9,
          tol=1e-4,
      ),
      dict(
          testcase_name='int16_quantized',
          recipe_path='../../recipes/default_a16w8_recipe.json',
          tensor_type=7,
          tol=2.5,  # TODO(b/379757798): Update tolerance after bug is fixed.
      ),
  )
  def test_sum_model_full_integer(self, recipe_path, tensor_type, tol):
    recipe_path = test_utils.get_path_to_datafile(recipe_path)
    self._quantizer.load_quantization_recipe(recipe_path)
    self.assertTrue(self._quantizer.need_calibration)

    data = _get_calibration_data()
    calibration_result = self._quantizer.calibrate(data)

    quantization_result = self._quantizer.quantize(calibration_result)

    # Check input/output tensor type.
    quantized_model = tfl_flatbuffer_utils.read_model(
        quantization_result.quantized_model
    )
    self.assertLen(quantized_model.subgraphs, 1)
    subgraph = quantized_model.subgraphs[0]
    subgraph_tensors = subgraph.tensors
    self.assertLen(subgraph.inputs, 1)
    input_tensor = subgraph_tensors[subgraph.inputs[0]]
    output_tensor = subgraph_tensors[subgraph.outputs[0]]
    # See schema_py_generated.py for type code.
    self.assertEqual(input_tensor.type, tensor_type)
    self.assertEqual(output_tensor.type, tensor_type)

    comparison_result = self._quantizer.validate(
        error_metrics='mse',
        test_data=_get_test_data(num_samples=1),
    )
    self._check_comparison_result(comparison_result, output_tolerance=tol)

  def _check_comparison_result(self, comparison_result, output_tolerance):
    # TODO: b/357959309 - Use comparison result directly for testing.
    comparison_result = comparison_result.get_all_tensor_results()
    output_mse = comparison_result['PartitionedCall:0']
    self.assertLess(output_mse, output_tolerance)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Sample end-to-end example for the MNIST toy model.

This script quantizes an MNIST toy model and runs inference on a sample MNIST
image.
"""

import random
from typing import Any, Optional

from absl import app
from absl import flags
import numpy as np
from PIL import Image
import tensorflow as tf

import os
import io
from ai_edge_quantizer import qtyping
from ai_edge_quantizer import quantizer
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_interpreter_utils
from ai_edge_litert import interpreter as tfl_interpreter  # pylint: disable=g-direct-tensorflow-import

_OpExecutionMode = qtyping.OpExecutionMode
_OpName = qtyping.TFLOperationName
_TensorQuantConfig = qtyping.TensorQuantizationConfig
_OpQuantConfig = qtyping.OpQuantizationConfig
_QuantGranularity = qtyping.QuantGranularity

_FLOAT_MODEL_PATH = flags.DEFINE_string(
    'float_model_path',
    test_utils.get_path_to_datafile('../../tests/models/conv_fc_mnist.tflite'),
    'The trained floating point MNIST toy model TFLite flatbuffer path.',
)
_RECIPE_PATH = flags.DEFINE_string(
    'recipe_path',
    test_utils.get_path_to_datafile(
        '../../recipes/default_af32w8float_recipe.json'
    ),
    'The quantization recipe path in JSON format.',
)
_IMG_PATH = flags.DEFINE_string(
    'img_path',
    test_utils.get_path_to_datafile('data/sample6.png'),
    'Path to the MNIST image to be predicted.',
)
_OUTPUT_DIR = flags.DEFINE_string(
    'output_dir',
    '/tmp/',
    'Directory to save the quantized model and recipe.',
)


def _check_user_inputs() -> None:
  """Check if the user input paths exist and create the save path if necessary."""
  if not os.path.exists(_FLOAT_MODEL_PATH.value):
    raise ValueError(
        'Model file does not exist. Please check the .tflite model path.'
    )
  if not os.path.exists(_RECIPE_PATH.value):
    raise ValueError(
        'Recipe file does not exist. Please check the recipe path.'
    )
  if not os.path.exists(_IMG_PATH.value):
    raise ValueError('Image file does not exist. Please check the image path.')

  if not os.path.exists(_OUTPUT_DIR.value):
    os.makedirs(_OUTPUT_DIR.value)


def quantize(
    float_model_path: str,
    recipe_json_path: str,
) -> Optional[bytearray]:
  """Quantize the float model.

  Args:
      float_model_path: Path to the float model.
      recipe_json_path: Path to the quantization recipe.

  Returns:
      Quantized model in bytes.
  """

  def _get_calibration_data(
      num_samples: int = 256,
  ) -> dict[str, list[dict[str, Any]]]:
    """Generate random dummy calibration data.

    The calibration data is a list of dictionaries, each of which contains an
    input data for a single calibration step. The key is the input tensor name
    and the value is the input tensor value.

    Args:
      num_samples: Number of samples to generate.

    Returns:
      A list of calibration data.
    """
    (x_train, _), _ = tf.keras.datasets.mnist.load_data()
    x_train = x_train / 255.0  # Normalize pixel values to 0-1.
    x_train = x_train.astype(np.float32)
    x_train = x_train.reshape([-1, 28, 28, 1])
    calibration_samples = []
    for _ in range(num_samples):
      sample = random.choice(x_train)
      calibration_samples.append(
          {'conv2d_input': sample.reshape([-1, 28, 28, 1])}
      )
    calibration_data = {
        tfl_interpreter_utils.DEFAULT_SIGNATURE_KEY: calibration_samples,
    }
    return calibration_data

  # 1) Instantiate a new quantizer with the source float model path.
  qt = quantizer.Quantizer(float_model_path)

  # 2) Initialize the quantization recipe. The input is a json file that
  #    specifies the quantization configuration for the model.
  #    A collection of common recipes can be found in the top-level directory
  #    `recipes/`. See `recipes/sample_advanced_usage_recipe.json` for
  #    advanced usage.
  #    Alternatively, you can also use the `update_quantization_recipe()`
  #    method to specify the quantization recipe programmatically.
  qt.load_quantization_recipe(recipe_json_path)

  # 3) Calibrate the model if necessary. This is only necessary for quantization
  #    modes that involve integer computation. `qt.need_calibration` can be used
  #    to check if the quantization recipe needs calibration.
  calibration_result = (
      qt.calibrate(_get_calibration_data()) if qt.need_calibration else None
  )

  # 4) Quantize the model. The quantization result contains the actual
  #    recipe and the quantized model.
  quant_result = qt.quantize(calibration_result)

  # 5) Save the quantized model and the recipe used to the filesystem.
  quant_result.save(_OUTPUT_DIR.value, model_name='mnist_toy_model')

  return quant_result.quantized_model


def inference(quantized_tflite: bytes, image_path: str):
  """Run inference on the quantized model.

  Args:
      quantized_tflite: Quantized model in bytes.
      image_path: Path to the image to be predicted.
  """

  def _read_img(img_path: str) -> np.ndarray:
    """Read MNIST image and normalize it.

    Args:
        img_path: Path to the image to be predicted.

    Returns:
        The normalized image.
    """
    image = Image.open(img_path)
    data = np.asarray(image, dtype=np.float32)
    if data.shape not in [(28, 28), (28, 28, 1)]:
      raise ValueError(
          'Invalid input image shape (MNIST image should have shape 28*28 or'
          ' 28*28*1)'
      )
    # Normalize the image if necessary.
    if data.max() > 1:
      data = data / 255.0
    return data.reshape((1, 28, 28, 1))

  tflite_interpreter = tfl_interpreter.Interpreter(
      model_content=quantized_tflite
  )
  tflite_interpreter.allocate_tensors()
  data = _read_img(image_path)
  tfl_interpreter_utils.invoke_interpreter_once(tflite_interpreter, [data])
  tflite_output_details = tflite_interpreter.get_output_details()[0]
  result = tflite_interpreter.get_tensor(tflite_output_details['index'])
  category_probabilities = np.squeeze(result)
  predicted_category = np.argmax(category_probabilities)
  print(
      f'Model predicts the image as {predicted_category} with probability'
      f' {category_probabilities[predicted_category]}',
  )


def main(_) -> None:
  _check_user_inputs()

  # Quantize the source floating point model into a quantized model based on the
  # given quantization recipe. This function demonstrates the end-to-end
  # quantization flow and the basic usage of AI Edge Quantizer's API.
  quantized_model = quantize(_FLOAT_MODEL_PATH.value, _RECIPE_PATH.value)

  # Run inference on the quantized model. This function demonstrates how to
  # use the quantized model for inference.
  inference(quantized_model, _IMG_PATH.value)


if __name__ == '__main__':
  app.run(main)



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from collections.abc import Generator
import pathlib
from typing import Any

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from ai_edge_quantizer import calibrator
from ai_edge_quantizer import params_generator
from ai_edge_quantizer import qtyping
from ai_edge_quantizer import recipe_manager
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils
from ai_edge_quantizer.utils import tfl_interpreter_utils


_ComputePrecision = qtyping.ComputePrecision
_TensorDataType = qtyping.TensorDataType
_TensorQuantConfig = qtyping.TensorQuantizationConfig
_QuantTransformation = qtyping.QuantTransformation
_AlgorithmName = recipe_manager.AlgorithmName
_QuantGranularity = qtyping.QuantGranularity
_QTransf = qtyping.QuantTransformation


TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile('')
_PARAMS_8BIT = qtyping.UniformQuantParams(8, None, np.array([1]), np.array([0]))


def _single_fc_model_representative_dataset_gen(num_samples=5):
  for _ in range(num_samples):
    yield {'input_1': np.random.rand(1, 8).astype(np.float32)}


def _int_transpose_model_representative_dataset_gen(num_samples=5):
  data = []
  for _ in range(num_samples):
    data.append({'input_2': np.random.rand(1, 2, 3, 4).astype(np.int32)})
  return data


def _get_calibration_data(
    dataset_gen: Generator[dict[str, Any], Any, None],
) -> dict[str, Any]:
  calibration_samples = [sample for sample in dataset_gen]
  calibration_data = {
      tfl_interpreter_utils.DEFAULT_SIGNATURE_KEY: calibration_samples,
  }
  return calibration_data


def _get_test_consumers(
    transformations_per_consumer: list[list[_QTransf]],
    params_per_consumer: list[qtyping.OpToTensorParams],
) -> list[qtyping.OpToTensorParams]:
  return [
      qtyping.OpToTensorParams(
          subgraph_op_id=i + 1,
          transformations=transformations_per_consumer[i],
          parameters=params_per_consumer[i],
      )
      for i in range(len(transformations_per_consumer))
  ]


class ParamsGeneratorTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'tests/models/conv_fc_mnist.tflite'
    )
    self._test_model = tfl_flatbuffer_utils.read_model(self._test_model_path)
    self._params_generator = params_generator.ParamsGenerator(self._test_model)
    self._recipe_manager = recipe_manager.RecipeManager()

  def test_update_model_quant_results(self):
    params_from_target = qtyping.TensorTransformationParams(
        tensor_name='test_tensor0',
        consumers=[
            qtyping.OpToTensorParams(
                subgraph_op_id=3,
                transformations=[
                    _QuantTransformation.ADD_QUANTIZE,
                    _QuantTransformation.ADD_DEQUANTIZE,
                ],
            )
        ],
    )
    # Test add new tensor from target
    self._params_generator._update_model_quant_results([params_from_target])
    self.assertIsNotNone(
        'test_tensor0' in self._params_generator.model_quant_results
    )
    tensor_params = self._params_generator.model_quant_results['test_tensor0']
    self.assertEqual(
        tensor_params,
        params_from_target,
    )
    # Test update new tensor from source
    params_from_source = qtyping.TensorTransformationParams(
        tensor_name='test_tensor0',
        producer=qtyping.OpToTensorParams(
            subgraph_op_id=3,
            transformations=[
                _QuantTransformation.ADD_DEQUANTIZE,
            ],
        ),
    )
    self._params_generator._update_model_quant_results([params_from_source])
    tensor_params = self._params_generator.model_quant_results['test_tensor0']
    self.assertEqual(
        tensor_params.producer,
        params_from_source.producer,
    )

    # We can have multiple target op params
    params_from_target2 = qtyping.TensorTransformationParams(
        tensor_name='test_tensor0',
        consumers=[
            qtyping.OpToTensorParams(
                subgraph_op_id=3,
                transformations=[
                    _QuantTransformation.NO_QUANTIZE,
                ],
            )
        ],
    )
    self._params_generator._update_model_quant_results([params_from_target2])
    tensor_params = self._params_generator.model_quant_results['test_tensor0']
    self.assertSequenceEqual(
        tensor_params.consumers,
        params_from_target.consumers + params_from_target2.consumers,
    )

    # but only a single source op params
    error_message = (
        'received multiple quantization parameters from the source op'
    )
    with self.assertRaisesWithPredicateMatch(
        RuntimeError, lambda err: error_message in str(err)
    ):
      self._params_generator._update_model_quant_results([params_from_source])

  def test_generate_config_global(self):
    # Quantize all fully_connected.
    global_recipe = [
        {
            'regex': '.*',
            'operation': 'FULLY_CONNECTED',
            'algorithm_key': 'min_max_uniform_quantize',
            'op_config': {
                'weight_tensor_config': {
                    'dtype': _TensorDataType.INT,
                    'num_bits': 8,
                    'symmetric': False,
                    'granularity': _QuantGranularity.CHANNELWISE,
                },
                # Equivalent to WEIGHT_ONLY.
                'compute_precision': _ComputePrecision.FLOAT,
                'explicit_dequantize': True,
            },
        },
    ]
    self._recipe_manager.load_quantization_recipe(global_recipe)
    tensor_quantization_params = (
        self._params_generator.generate_quantization_parameters(
            self._recipe_manager
        )
    )
    # Every tensor in the model will have their params!
    flatbuffer_model = tfl_flatbuffer_utils.read_model(self._test_model_path)
    tensors = flatbuffer_model.subgraphs[0].tensors
    self.assertLen(tensor_quantization_params, len(tensors))

    # Input tensor
    tensor_name = 'serving_default_conv2d_input:0'
    self._test_tensor_transformation_params(
        0,
        tensor_quantization_params,
        tensor_name,
        [_QuantTransformation.NO_QUANTIZE],
        is_inbounding_tensor=True,
    )
    # Input tensor is produced from the virtual Input op.
    transformation_config = tensor_quantization_params[tensor_name]
    self.assertIsNotNone(transformation_config.producer)
    self.assertEqual(transformation_config.producer.subgraph_op_id, -1)

    # Intermediate tensor will have no_quantize at the both end
    tensor_name = 'sequential/average_pooling2d/AvgPool'
    self._test_tensor_transformation_params(
        1,
        tensor_quantization_params,
        tensor_name,
        [_QuantTransformation.NO_QUANTIZE],
        is_inbounding_tensor=False,
    )  # output from average pool
    self._test_tensor_transformation_params(
        2,
        tensor_quantization_params,
        tensor_name,
        [_QuantTransformation.NO_QUANTIZE],
        is_inbounding_tensor=True,
    )  # input to Reshape

    # First FC
    self._test_tensor_transformation_params(
        3,
        tensor_quantization_params,
        'sequential/flatten/Reshape',
        [_QuantTransformation.NO_QUANTIZE],
        is_inbounding_tensor=True,
    )  # input tensor

    self._test_tensor_transformation_params(
        3,
        tensor_quantization_params,
        'arith.constant1',
        [
            _QuantTransformation.ADD_DEQUANTIZE,
        ],
        is_inbounding_tensor=True,
        num_bits=8,
        granularity=_QuantGranularity.CHANNELWISE,
        symmetric=False,
    )  # weight tensor
    self._test_tensor_transformation_params(
        3,
        tensor_quantization_params,
        'arith.constant2',
        [_QuantTransformation.NO_QUANTIZE],
        is_inbounding_tensor=True,
    )  # bias tensor
    self._test_tensor_transformation_params(
        3,
        tensor_quantization_params,
        'sequential/dense/MatMul;sequential/dense/Relu;sequential/dense/BiasAdd',
        [_QuantTransformation.NO_QUANTIZE],
        is_inbounding_tensor=False,
    )  # output tensor

    # Second FC
    self._test_tensor_transformation_params(
        4,
        tensor_quantization_params,
        'sequential/dense/MatMul;sequential/dense/Relu;sequential/dense/BiasAdd',
        [_QuantTransformation.NO_QUANTIZE],
        is_inbounding_tensor=True,
    )  # input tensor
    self._test_tensor_transformation_params(
        4,
        tensor_quantization_params,
        'arith.constant',
        [
            _QuantTransformation.ADD_DEQUANTIZE,
        ],
        is_inbounding_tensor=True,
        num_bits=8,
        granularity=_QuantGranularity.CHANNELWISE,
        symmetric=False,
    )  # weight tensor
    self._test_tensor_transformation_params(
        4,
        tensor_quantization_params,
        'sequential/dense_1/MatMul',
        [_QuantTransformation.NO_QUANTIZE],
        is_inbounding_tensor=False,
    )  # output tensor

    # Model output tensor
    tensor_name = 'StatefulPartitionedCall:0'
    self._test_tensor_transformation_params(
        5,
        tensor_quantization_params,
        tensor_name,
        [_QuantTransformation.NO_QUANTIZE],
        is_inbounding_tensor=False,
    )
    # Output tensor is consumed by the virtual Output op.
    transformation_config = tensor_quantization_params[tensor_name]
    self.assertLen(transformation_config.consumers, 1)
    consumer = transformation_config.consumers[0]
    self.assertEqual(consumer.subgraph_op_id, -1)

  # TODO: b/330770656 - expand the test to cover mixed activation precision.
  def test_generate_config_selective(self):
    # Choose scope regex using Model Explorer
    selective_quantization_recipe = [
        {
            'regex': '.*/dense/.*',
            'operation': 'FULLY_CONNECTED',
            'algorithm_key': 'min_max_uniform_quantize',
            'op_config': {
                'weight_tensor_config': {
                    'dtype': _TensorDataType.INT,
                    'num_bits': 8,
                    'symmetric': True,
                    'granularity': _QuantGranularity.CHANNELWISE,
                },
                # Equivalent to DRQ.
                'compute_precision': _ComputePrecision.INTEGER,
                'explicit_dequantize': False,
            },
        },
        {
            'regex': '.*/dense_1/.*',
            'operation': 'FULLY_CONNECTED',
            'algorithm_key': 'min_max_uniform_quantize',
            'op_config': {
                'weight_tensor_config': {
                    'dtype': _TensorDataType.INT,
                    'num_bits': 4,
                    'symmetric': False,
                    'granularity': _QuantGranularity.TENSORWISE,
                },
                # Equivalent to WEIGHT_ONLY.
                'compute_precision': _ComputePrecision.FLOAT,
                'explicit_dequantize': True,
            },
        },
    ]
    self._recipe_manager.load_quantization_recipe(selective_quantization_recipe)
    tensor_quantization_params = (
        self._params_generator.generate_quantization_parameters(
            self._recipe_manager
        )
    )
    # FC weights for scope "dense"
    self._test_tensor_transformation_params(
        3,
        tensor_quantization_params,
        'arith.constant1',
        [_QuantTransformation.QUANTIZE_TENSOR],
        is_inbounding_tensor=True,
        num_bits=8,
        granularity=_QuantGranularity.CHANNELWISE,
        symmetric=True,
    )

    # FC weights for scope "dense1"
    self._test_tensor_transformation_params(
        4,
        tensor_quantization_params,
        'arith.constant',
        [
            _QuantTransformation.ADD_DEQUANTIZE,
        ],
        is_inbounding_tensor=True,
        num_bits=4,
        granularity=_QuantGranularity.TENSORWISE,
        symmetric=False,
    )

  def test_generate_config_edge_cases(self):

    selective_quantization_recipe = [
        # Use the tensor name as scope directly.
        {
            'regex': 'sequential/dense_1/MatMul',
            'operation': 'FULLY_CONNECTED',
            'algorithm_key': 'min_max_uniform_quantize',
            'op_config': {
                'weight_tensor_config': {
                    'num_bits': 8,
                    'symmetric': True,
                    'granularity': _QuantGranularity.CHANNELWISE,
                },
                # Equivalent to DRQ.
                'compute_precision': _ComputePrecision.INTEGER,
            },
        },
        # Scope that does not exist in the model.
        {
            'regex': '.*/dense_3/.*',
            'operation': 'FULLY_CONNECTED',
            'algorithm_key': 'min_max_uniform_quantize',
            'op_config': {
                'weight_tensor_config': {
                    'num_bits': 4,
                    'symmetric': False,
                    'granularity': _QuantGranularity.TENSORWISE,
                },
                # Equivalent to WEIGHT_ONLY.
                'compute_precision': _ComputePrecision.FLOAT,
                'explicit_dequantize': True,
            },
        },
    ]
    self._recipe_manager.load_quantization_recipe(selective_quantization_recipe)
    tensor_quantization_params = (
        self._params_generator.generate_quantization_parameters(
            self._recipe_manager
        )
    )
    # Only the second FC will be quantized
    self._test_tensor_transformation_params(
        3,
        tensor_quantization_params,
        'arith.constant1',
        [_QuantTransformation.NO_QUANTIZE],
        is_inbounding_tensor=True,
    )

    self._test_tensor_transformation_params(
        4,
        tensor_quantization_params,
        'arith.constant',
        [_QuantTransformation.QUANTIZE_TENSOR],
        is_inbounding_tensor=True,
        num_bits=8,
        granularity=_QuantGranularity.CHANNELWISE,
        symmetric=True,
    )

  @parameterized.parameters(
      (True, _QuantGranularity.CHANNELWISE),
      (True, _QuantGranularity.TENSORWISE),
      (False, _QuantGranularity.CHANNELWISE),
      (False, _QuantGranularity.TENSORWISE),
  )
  def test_generate_config_int8xint8_single_fc(
      self, act_symmetric, channelwise_weight
  ):
    single_fc_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH) / 'tests/models/single_fc.tflite'
    )
    single_fc_model = tfl_flatbuffer_utils.read_model(single_fc_model_path)
    params_generator_single_fc = params_generator.ParamsGenerator(
        single_fc_model
    )
    self._recipe_manager.add_quantization_config(
        regex='.*',
        operation_name=qtyping.TFLOperationName.ALL_SUPPORTED,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=qtyping.OpQuantizationConfig(
            activation_tensor_config=_TensorQuantConfig(
                num_bits=8, symmetric=act_symmetric
            ),
            weight_tensor_config=_TensorQuantConfig(
                num_bits=8, symmetric=True, granularity=channelwise_weight
            ),
            # Equivalent to SRQ.
            compute_precision=_ComputePrecision.INTEGER,
        ),
    )

    # Raise error when missing QSVs.
    error_message = 'Model quantization statistics values (QSVs) are required'
    with self.assertRaisesWithPredicateMatch(
        RuntimeError, lambda err: error_message in str(err)
    ):
      params_generator_single_fc.generate_quantization_parameters(
          self._recipe_manager
      )

    # Calibrate then quantize
    model_calibrator = calibrator.Calibrator(single_fc_model_path)
    calibration_data = _get_calibration_data(
        _single_fc_model_representative_dataset_gen()
    )
    model_calibrator.calibrate(calibration_data, self._recipe_manager)
    model_qsvs = model_calibrator.get_model_qsvs()
    quant_params = params_generator_single_fc.generate_quantization_parameters(
        self._recipe_manager,
        model_qsvs,
    )
    self.assertLen(quant_params, 4)

    # Input tensor producer (from the virtual input op).
    self._test_tensor_transformation_params(
        -1,  # virtual input op.
        quant_params,
        'serving_default_input_1:0',
        [_QuantTransformation.ADD_DEQUANTIZE],
        num_bits=8,
        granularity=_QuantGranularity.TENSORWISE,
        symmetric=act_symmetric,
        is_inbounding_tensor=False,
    )
    # Input tensor consumer.
    self._test_tensor_transformation_params(
        0,
        quant_params,
        'serving_default_input_1:0',
        [_QuantTransformation.ADD_QUANTIZE],
        num_bits=8,
        granularity=_QuantGranularity.TENSORWISE,
        symmetric=act_symmetric,
        is_inbounding_tensor=True,
    )

    # output tensor producer.
    self._test_tensor_transformation_params(
        0,
        quant_params,
        'StatefulPartitionedCall:0',
        [_QuantTransformation.ADD_DEQUANTIZE],
        num_bits=8,
        granularity=_QuantGranularity.TENSORWISE,
        symmetric=act_symmetric,
        is_inbounding_tensor=False,
    )
    # output tensor consumer (into the virtual output op).
    self._test_tensor_transformation_params(
        -1,  # virtual output op.
        quant_params,
        'StatefulPartitionedCall:0',
        [_QuantTransformation.ADD_QUANTIZE],
        num_bits=8,
        granularity=_QuantGranularity.TENSORWISE,
        symmetric=act_symmetric,
        is_inbounding_tensor=True,
    )

    # weights
    self._test_tensor_transformation_params(
        0,
        quant_params,
        'sequential/dense/MatMul',
        [_QuantTransformation.QUANTIZE_TENSOR],
        num_bits=8,
        granularity=channelwise_weight,
        symmetric=True,
        is_inbounding_tensor=True,
    )

    # bias
    self._test_tensor_transformation_params(
        0,
        quant_params,
        'sequential/dense/BiasAdd/ReadVariableOp',
        [_QuantTransformation.QUANTIZE_TENSOR],
        num_bits=32,
        granularity=channelwise_weight,
        symmetric=True,
        is_inbounding_tensor=True,
    )

  @parameterized.parameters('weight_only', 'DRQ')
  def test_generate_params_buffer_sharing_graphs_succeeds(
      self, the_other_fc_difference
  ):
    model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'tests/models/weight_sharing_fcs.tflite'
    )
    model = tfl_flatbuffer_utils.read_model(model_path)
    pg = params_generator.ParamsGenerator(model)
    self._recipe_manager.add_quantization_config(
        regex='.*',
        operation_name=qtyping.TFLOperationName.ALL_SUPPORTED,
        op_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(num_bits=8, symmetric=True),
            # Equivalent to WEIGHT_ONLY.
            compute_precision=_ComputePrecision.FLOAT,
            explicit_dequantize=True,
        ),
    )
    if the_other_fc_difference == 'DRQ':
      self._recipe_manager.add_quantization_config(
          regex='PartitionedCall_1:0',
          operation_name=qtyping.TFLOperationName.ALL_SUPPORTED,
          op_config=qtyping.OpQuantizationConfig(
              # Equivalent to DRQ.
              compute_precision=_ComputePrecision.INTEGER,
          ),
      )
    quant_params = pg.generate_quantization_parameters(
        self._recipe_manager,
    )
    self.assertLen(quant_params, 6)

  @parameterized.named_parameters(
      dict(
          testcase_name='different_quant_config_fc2_no_quant',
          fc_2_num_bits=None,
          expected_tensor_with_buffer_duplication='BatchMatMulV3',
      ),
      dict(
          testcase_name='different_quant_config_fc2_int4',
          fc_2_num_bits=4,
          expected_tensor_with_buffer_duplication='BatchMatMulV3',
      ),
      dict(
          testcase_name='same_quant_config',
          fc_2_num_bits=8,
          expected_tensor_with_buffer_duplication=None,
      ),
  )
  def test_generate_params_marks_correct_buffers_for_duplication_when_distinct_tensors_share_constant_buffer(
      self,
      fc_2_num_bits,
      expected_tensor_with_buffer_duplication,
  ):
    model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'tests/models/weight_sharing_fcs.tflite'
    )
    model = tfl_flatbuffer_utils.read_model(model_path)
    pg = params_generator.ParamsGenerator(model)

    # Setup the quantization config for the first FC.
    self._recipe_manager.add_quantization_config(
        regex='PartitionedCall:0',
        operation_name=qtyping.TFLOperationName.FULLY_CONNECTED,
        op_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(
                num_bits=8, granularity=qtyping.QuantGranularity.CHANNELWISE
            ),
            compute_precision=_ComputePrecision.INTEGER,
        ),
    )
    # Setup the quantization config for the second FC (weight shared with the
    # first FC).
    if fc_2_num_bits is not None:
      self._recipe_manager.add_quantization_config(
          regex='PartitionedCall_1:0',
          operation_name=qtyping.TFLOperationName.FULLY_CONNECTED,
          op_config=qtyping.OpQuantizationConfig(
              weight_tensor_config=_TensorQuantConfig(
                  num_bits=fc_2_num_bits,
                  granularity=qtyping.QuantGranularity.CHANNELWISE,
              ),
              compute_precision=_ComputePrecision.INTEGER,
          ),
      )

    quant_params = pg.generate_quantization_parameters(
        self._recipe_manager,
    )
    self.assertLen(quant_params, 6)

    # Check that the expected tensor has buffer duplication transformation as
    # the first one to be applied. And no other tensor has buffer duplication
    # transformation at all.
    for tensor_name in quant_params:
      if tensor_name == expected_tensor_with_buffer_duplication:
        self.assertIsNotNone(quant_params[tensor_name].consumers)
        for consumer in quant_params[tensor_name].consumers:
          self.assertNotEmpty(consumer.transformations)
          self.assertEqual(
              consumer.transformations[0],
              _QTransf.DUPLICATE_BUFFER,
          )
          self.assertNotIn(
              _QTransf.DUPLICATE_BUFFER, consumer.transformations[1:]
          )
      elif quant_params[tensor_name].consumers is not None:
        for consumer in quant_params[tensor_name].consumers:
          self.assertNotIn(_QTransf.DUPLICATE_BUFFER, consumer.transformations)

  def _get_fc_recipe_entry(self, regex: str, num_bits: int):
    return {
        'regex': regex,
        'operation': 'FULLY_CONNECTED',
        'algorithm_key': 'min_max_uniform_quantize',
        'op_config': {
            'weight_tensor_config': {
                'num_bits': num_bits,
                'symmetric': True,
                'granularity': 'CHANNELWISE',
                'dtype': 'INT',
                'block_size': 0,
            },
            'compute_precision': 'INTEGER',
            'explicit_dequantize': False,
            'skip_checks': False,
            'min_weight_elements': 0,
        },
    }

  @parameterized.named_parameters(
      dict(
          testcase_name='fc1_quant_fc2_no_quant',
          fc1_num_bits=8,
          fc2_num_bits=None,
      ),
      dict(
          testcase_name='fc1_no_quant_fc2_quant',
          fc1_num_bits=None,
          fc2_num_bits=8,
      ),
      dict(
          testcase_name='fc1_quant_fc2_quant_different_params',
          fc1_num_bits=8,
          fc2_num_bits=4,
      ),
  )
  def test_generate_params_marks_correct_buffers_tensors_for_duplication(
      self,
      fc1_num_bits,
      fc2_num_bits,
  ):
    model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'tests/models/constant_tensor_and_buffer_only_sharing_weight_fcs.tflite',
    )
    model = tfl_flatbuffer_utils.read_model(model_path)
    pg = params_generator.ParamsGenerator(model)
    
    sig1_fc1_regex = 'BatchMatMulV3;'
    sig1_fc2_regex = 'PartitionedCall:0;'
    recipe = []
    if fc1_num_bits is not None:
      recipe.append(self._get_fc_recipe_entry(sig1_fc1_regex, fc1_num_bits))
    if fc2_num_bits is not None:
      recipe.append(self._get_fc_recipe_entry(sig1_fc2_regex, fc2_num_bits))
    self._recipe_manager.load_quantization_recipe(recipe)
    quant_params = pg.generate_quantization_parameters(self._recipe_manager)

    expected_tensor = 'arith.constant'
    consumers = quant_params[expected_tensor].consumers
    self.assertLen(consumers, 2)

    # Check FC1 transformations.
    if fc1_num_bits is None:
      fc1_quant_transformation = _QTransf.NO_QUANTIZE
    else:
      fc1_quant_transformation = _QTransf.QUANTIZE_TENSOR
    self.assertEqual(
        consumers[0].transformations,
        [
            _QTransf.DUPLICATE_TENSOR,
            _QTransf.DUPLICATE_BUFFER,
            fc1_quant_transformation,
        ],
    )
    # Check FC2 transformations.
    if fc2_num_bits is None:
      fc2_quant_transformation = _QTransf.NO_QUANTIZE
    else:
      fc2_quant_transformation = _QTransf.QUANTIZE_TENSOR
    self.assertEqual(
        consumers[1].transformations,
        [
            _QTransf.DUPLICATE_TENSOR,
            _QTransf.DUPLICATE_BUFFER,
            fc2_quant_transformation,
        ],
    )
    # Check that no other tensor has tensor or buffer duplication
    # transformations.
    for tensor_name, params in quant_params.items():
      if tensor_name == expected_tensor:
        continue
      for consumer in params.consumers:
        self.assertNotIn(_QTransf.DUPLICATE_TENSOR, consumer.transformations)
        self.assertNotIn(_QTransf.DUPLICATE_BUFFER, consumer.transformations)

  def test_generate_params_returns_valid_results_when_multiple_tensor_duplication_for_one_buffer(
      self,
  ):
    model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'tests/models/constant_tensor_and_buffer_only_sharing_weight_fcs.tflite',
    )
    model = tfl_flatbuffer_utils.read_model(model_path)
    pg = params_generator.ParamsGenerator(model)

    sig1_fc1_regex = 'BatchMatMulV3;'
    sig1_fc2_regex = 'PartitionedCall:0;'
    sig2_fc1_regex = 'BatchMatMulV31;'
    sig2_fc2_regex = 'PartitionedCall_1:0;'
    recipe = [
        self._get_fc_recipe_entry(sig1_fc1_regex, num_bits=8),
        self._get_fc_recipe_entry(sig1_fc2_regex, num_bits=4),
        self._get_fc_recipe_entry(sig2_fc1_regex, num_bits=8),
        self._get_fc_recipe_entry(sig2_fc2_regex, num_bits=4),
    ]
    self._recipe_manager.load_quantization_recipe(recipe)
    quant_params = pg.generate_quantization_parameters(self._recipe_manager)

    # Check transformations for sig1.
    sig1_expected_tensor = 'arith.constant'
    sig1_consumers = quant_params[sig1_expected_tensor].consumers
    self.assertLen(sig1_consumers, 2)
    sig1_expected_transformations = [
        _QTransf.DUPLICATE_TENSOR,
        _QTransf.DUPLICATE_BUFFER,
        _QTransf.QUANTIZE_TENSOR,
    ]
    for sig1_consumer in sig1_consumers:
      self.assertEqual(
          sig1_consumer.transformations,
          sig1_expected_transformations,
      )
    # Check transformations for sig2.
    sig2_expected_tensor = 'arith.constant1'
    sig2_consumers = quant_params[sig2_expected_tensor].consumers
    self.assertLen(sig2_consumers, 2)
    sig2_expected_transformations = [
        _QTransf.DUPLICATE_TENSOR,
        _QTransf.QUANTIZE_TENSOR,
    ]
    for sig2_consumer in sig2_consumers:
      self.assertEqual(
          sig2_consumer.transformations,
          sig2_expected_transformations,
      )
    # Check that no other tensor has tensor or buffer duplication
    # transformations.
    for tensor_name, params in quant_params.items():
      if tensor_name in [sig1_expected_tensor, sig2_expected_tensor]:
        continue
      for consumer in params.consumers:
        self.assertNotIn(_QTransf.DUPLICATE_TENSOR, consumer.transformations)
        self.assertNotIn(_QTransf.DUPLICATE_BUFFER, consumer.transformations)

  @parameterized.named_parameters(
      dict(
          testcase_name='producer_incompatible',
          param1=qtyping.TensorTransformationParams(
              tensor_name='tfl.quantize',
              producer=qtyping.OpToTensorParams(
                  subgraph_op_id=0,
                  transformations=[_QTransf.ADD_DEQUANTIZE],
                  parameters=_PARAMS_8BIT,
              ),
              consumers=_get_test_consumers(
                  transformations_per_consumer=[
                      [_QTransf.ADD_QUANTIZE],
                      [_QTransf.ADD_QUANTIZE, _QTransf.ADD_DEQUANTIZE],
                      [_QTransf.ADD_QUANTIZE, _QTransf.ADD_DEQUANTIZE],
                      [_QTransf.NO_QUANTIZE],
                  ],
                  params_per_consumer=[_PARAMS_8BIT] * 4,
              ),
          ),
          param2=qtyping.TensorTransformationParams(
              tensor_name='tfl.other_quantize',
              producer=qtyping.OpToTensorParams(
                  subgraph_op_id=0,
                  transformations=[_QTransf.NO_QUANTIZE],
                  parameters=_PARAMS_8BIT,
              ),
              consumers=_get_test_consumers(
                  transformations_per_consumer=[
                      [_QTransf.ADD_QUANTIZE],
                      [_QTransf.ADD_QUANTIZE, _QTransf.ADD_DEQUANTIZE],
                      [_QTransf.ADD_QUANTIZE, _QTransf.ADD_DEQUANTIZE],
                  ],
                  params_per_consumer=[_PARAMS_8BIT] * 4,
              ),
          ),
          expected=False,
      ),
      dict(
          testcase_name='compatible',
          param1=qtyping.TensorTransformationParams(
              tensor_name='tfl.quantize',
              producer=None,
              consumers=_get_test_consumers(
                  transformations_per_consumer=[
                      [_QTransf.ADD_QUANTIZE],
                      [_QTransf.NO_QUANTIZE, _QTransf.ADD_QUANTIZE],
                      [_QTransf.NO_QUANTIZE],
                  ],
                  params_per_consumer=[_PARAMS_8BIT] * 4,
              ),
          ),
          param2=qtyping.TensorTransformationParams(
              tensor_name='tfl.other_quantize',
              producer=None,
              consumers=_get_test_consumers(
                  transformations_per_consumer=[
                      [_QTransf.ADD_QUANTIZE],
                      [_QTransf.ADD_QUANTIZE, _QTransf.ADD_DEQUANTIZE],
                      [_QTransf.ADD_QUANTIZE, _QTransf.ADD_DEQUANTIZE],
                      [_QTransf.ADD_QUANTIZE],
                  ],
                  params_per_consumer=[_PARAMS_8BIT] * 4,
              ),
          ),
          expected=True,
      ),
      dict(
          testcase_name='compatible_no_numeric_check',
          param1=qtyping.TensorTransformationParams(
              tensor_name='tfl.quantize',
              producer=None,
              consumers=_get_test_consumers(
                  transformations_per_consumer=[
                      [_QTransf.ADD_QUANTIZE],
                      [_QTransf.ADD_QUANTIZE],
                  ],
                  params_per_consumer=[
                      qtyping.UniformQuantParams(
                          8, None, np.array([0.00028806]), np.array([0])
                      ),
                      qtyping.UniformQuantParams(
                          8, None, np.array([0.00027501]), np.array([0])
                      ),
                  ],
              ),
          ),
          param2=qtyping.TensorTransformationParams(
              tensor_name='tfl.quantize',
              producer=None,
              consumers=_get_test_consumers(
                  transformations_per_consumer=[
                      [_QTransf.ADD_QUANTIZE],
                      [_QTransf.ADD_QUANTIZE],
                  ],
                  params_per_consumer=[
                      qtyping.UniformQuantParams(
                          8, None, np.array([0.00028806]), np.array([0])
                      ),
                      qtyping.UniformQuantParams(
                          8, None, np.array([0.00027501]), np.array([0])
                      ),
                  ],
              ),
          ),
          expected=True,
      ),
  )
  def test__are_self_compatible_tensors_compatible_to_each_other(
      self, param1, param2, expected
  ):
    self.assertEqual(
        params_generator._are_self_compatible_tensors_compatible_to_each_other(
            param1, param2
        ),
        expected,
    )

  @parameterized.named_parameters(
      dict(
          testcase_name='consumer_incompatible',
          params=qtyping.TensorTransformationParams(
              tensor_name='tfl.quantize',
              producer=qtyping.OpToTensorParams(
                  subgraph_op_id=0,
                  transformations=[_QTransf.NO_QUANTIZE],
                  parameters=_PARAMS_8BIT,
              ),
              consumers=_get_test_consumers(
                  transformations_per_consumer=[
                      [_QTransf.ADD_QUANTIZE],
                      [_QTransf.ADD_QUANTIZE, _QTransf.ADD_DEQUANTIZE],
                      [_QTransf.ADD_QUANTIZE, _QTransf.ADD_DEQUANTIZE],
                      [_QTransf.QUANTIZE_TENSOR],
                  ],
                  params_per_consumer=[_PARAMS_8BIT] * 4,
              ),
          ),
          expected=False,
      ),
      dict(
          testcase_name='compatible',
          params=qtyping.TensorTransformationParams(
              tensor_name='tfl.quantize',
              producer=None,
              consumers=_get_test_consumers(
                  transformations_per_consumer=[
                      [_QTransf.ADD_QUANTIZE, _QTransf.ADD_DEQUANTIZE],
                      [_QTransf.ADD_QUANTIZE],
                      [_QTransf.NO_QUANTIZE, _QTransf.ADD_QUANTIZE],
                      [_QTransf.NO_QUANTIZE],
                  ],
                  params_per_consumer=[_PARAMS_8BIT] * 4,
              ),
          ),
          expected=True,
      ),
      dict(
          testcase_name='compatible_no_numeric_check',
          params=qtyping.TensorTransformationParams(
              tensor_name='tfl.quantize',
              producer=None,
              consumers=_get_test_consumers(
                  transformations_per_consumer=[
                      [_QTransf.ADD_QUANTIZE],
                      [_QTransf.ADD_QUANTIZE],
                  ],
                  params_per_consumer=[
                      qtyping.UniformQuantParams(
                          8, None, np.array([0.00028806]), np.array([0])
                      ),
                      qtyping.UniformQuantParams(
                          8, None, np.array([0.00027501]), np.array([0])
                      ),
                  ],
              ),
          ),
          expected=True,
      ),
  )
  def test__are_tensor_consumer_params_compatible(self, params, expected):
    self.assertEqual(
        params_generator._are_tensor_consumer_params_compatible(params),
        expected,
    )

  def test_model_with_duplicated_tensor_names_fails(self):
    model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'tests/models/duplicated_tensor_names.tflite'
    )
    model = tfl_flatbuffer_utils.read_model(model_path)
    error_message = 'Tensor name test_same_name is not unique in the model.'
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      params_generator.ParamsGenerator(model)

  def test_quantize_integer_input_output(self):
    model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'tests/models/single_transpose_int32.tflite'
    )
    model = tfl_flatbuffer_utils.read_model(model_path)
    self._recipe_manager.add_quantization_config(
        regex='.*',
        operation_name=qtyping.TFLOperationName.ALL_SUPPORTED,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=qtyping.OpQuantizationConfig(
            activation_tensor_config=_TensorQuantConfig(
                num_bits=8, symmetric=False
            ),
            weight_tensor_config=_TensorQuantConfig(num_bits=8, symmetric=True),
            # Equivalent to SRQ.
            compute_precision=_ComputePrecision.INTEGER,
        ),
    )
    pg = params_generator.ParamsGenerator(model)

    # Calibrate then quantize.
    model_calibrator = calibrator.Calibrator(model_path)
    calibration_data = _get_calibration_data(
        _int_transpose_model_representative_dataset_gen()
    )
    model_calibrator.calibrate(calibration_data, self._recipe_manager)
    model_qsvs = model_calibrator.get_model_qsvs()
    quant_params = pg.generate_quantization_parameters(
        self._recipe_manager,
        model_qsvs,
    )
    self.assertLen(quant_params, 3)

    self._test_tensor_transformation_params(
        -1,  # virtual input op.
        quant_params,
        'serving_default_input_2:0',
        [_QuantTransformation.NO_QUANTIZE],
        is_inbounding_tensor=False,
    )
    # Input tensor consumer.
    self._test_tensor_transformation_params(
        0,
        quant_params,
        'serving_default_input_2:0',
        [_QuantTransformation.NO_QUANTIZE],
        is_inbounding_tensor=True,
    )

    # Output tensor producer.
    self._test_tensor_transformation_params(
        0,
        quant_params,
        'PartitionedCall:0',
        [_QuantTransformation.NO_QUANTIZE],
        is_inbounding_tensor=False,
    )
    # output tensor consumer (into the virtual output op).
    self._test_tensor_transformation_params(
        -1,  # virtual output op.
        quant_params,
        'PartitionedCall:0',
        [_QuantTransformation.NO_QUANTIZE],
        is_inbounding_tensor=True,
    )

    # perm
    self._test_tensor_transformation_params(
        0,
        quant_params,
        'sequential_1/permute_1/transpose/perm',
        [_QuantTransformation.NO_QUANTIZE],
        is_inbounding_tensor=True,
    )

  def _test_tensor_transformation_params(
      self,
      subgraph_op_id,
      quant_params,
      tensor_name,
      transformations,
      is_inbounding_tensor,
      num_bits=8,
      granularity=_QuantGranularity.TENSORWISE,
      symmetric=True,
      quantized_dimension=0,
  ):
    """Helper function to test tensor transformation parameters are correct."""
    self.assertIn(tensor_name, quant_params)
    transformation_config = quant_params[tensor_name]
    self.assertEqual(transformation_config.tensor_name, tensor_name)
    if is_inbounding_tensor:
      self.assertLen(transformation_config.consumers, 1)
      op_config = transformation_config.consumers[0]
    else:
      op_config = transformation_config.producer
    self.assertIsNotNone(op_config)
    self.assertEqual(op_config.subgraph_op_id, subgraph_op_id)
    self.assertSequenceEqual(op_config.transformations, transformations)
    if transformations == [_QuantTransformation.NO_QUANTIZE]:
      self.assertIsNone(op_config.parameters)
    else:
      quantization_params = op_config.parameters
      self.assertIsNotNone(quantization_params)
      if granularity is _QuantGranularity.CHANNELWISE:
        self.assertEqual(
            quantization_params.quantized_dimension, quantized_dimension
        )
      else:
        self.assertIsNone(quantization_params.quantized_dimension)
      self.assertEqual(quantization_params.num_bits, num_bits)
      if symmetric:
        self.assertEqual(np.sum(abs(quantization_params.zero_point)), 0)
      else:
        self.assertEqual(
            len(quantization_params.scale),
            len(quantization_params.zero_point),
        )


class ParamsGeneratorAlreadyQuantizedModelTest(absltest.TestCase):

  def test_check_is_float_model_succeeds_when_model_is_float(self):
    test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'tests/models/conv_fc_mnist.tflite'
    )
    test_model = tfl_flatbuffer_utils.read_model(test_model_path)
    _ = params_generator.ParamsGenerator(test_model)

  def test_check_is_quantized_model_succeeds_when_model_is_quantized(self):
    test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'tests/models/mnist_quantized.tflite'
    )
    test_model = tfl_flatbuffer_utils.read_model(test_model_path)
    _ = params_generator.ParamsGenerator(test_model)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Test utils for uniform quantize op architecture tests."""

from collections.abc import Callable, Sequence
import dataclasses
from typing import Any, Optional

from absl.testing import parameterized
import numpy as np

from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.uniform_quantize import uniform_quantize_tensor
from ai_edge_quantizer.algorithms.utils import common_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils

_TFLOpName = qtyping.TFLOperationName
_ComputePrecision = qtyping.ComputePrecision
_TensorQuantConfig = qtyping.TensorQuantizationConfig
_QuantTransformation = qtyping.QuantTransformation


@dataclasses.dataclass
class OpTestInfo:
  """Aggregate op test information.

  Attributes:
    test_model: The test model.
    op_tensor_names: A map of op tensor names with keys for input(s) and output.
      Note for multiple inputs, the keys will follow the pattern of "input",
      "input2", "input3" etc. for ops tested with _test_no_weights_op. For ops
      tested with _test_fc_bmm_conv, the format for non-ignored inputs is the
      same, but the ignored tensors should have keys "ignored_input",
      "ignored_input2", "ignored_input3" etc.
    input_range: A tuple of input min and max.
    output_range: A tuple of output min and max.
    quantized_dimension: Quantized dimension.
  """

  test_model: Any
  op_tensor_names: dict[str, str]
  input_range: tuple[np.ndarray, np.ndarray]
  output_range: tuple[np.ndarray, np.ndarray]
  quantized_dimension: int = 0


DEFAULT_ACTIVATION_QUANT_SETTING = _TensorQuantConfig(
    num_bits=8,
    symmetric=False,
    granularity=qtyping.QuantGranularity.TENSORWISE,
)
DEFAULT_WEIGHT_QUANT_SETTING = _TensorQuantConfig(
    num_bits=8,
    symmetric=True,
    granularity=qtyping.QuantGranularity.CHANNELWISE,
)


class BaseQuantizeTest(parameterized.TestCase):
  """Base test class for uniform quantize op architecture tests.

  This class provides test utils for uniform quantize op architecture tests.
  """

  def setUp(self):
    super().setUp()
    self._tensor_name_to_qsv = {}

  def _setup_op_test_config(
      self,
      compute_precision,
      is_drq,
      op_test_info,
      num_inputs=1,
      num_outputs=1,
  ):
    """Helper to set up qsv for op test."""
    # SRQ requires QSVs (min/max).
    input_min, input_max = op_test_info.input_range
    output_min, output_max = op_test_info.output_range
    # Check if SRQ.
    if compute_precision == _ComputePrecision.INTEGER and not is_drq:
      input_qsv = {
          "min": input_min,
          "max": input_max,
      }
      output_qsv = {
          "min": output_min,
          "max": output_max,
      }
      for i in range(num_inputs):
        input_name = "input"
        if i > 0:
          input_name = f"input{i+1}"
        self._tensor_name_to_qsv[op_test_info.op_tensor_names[input_name]] = (
            input_qsv
        )
      for i in range(num_outputs):
        output_name = "output"
        if i > 0:
          output_name = f"output{i+1}"
        self._tensor_name_to_qsv[op_test_info.op_tensor_names[output_name]] = (
            output_qsv
        )

  def _test_same_direction_tensors(
      self,
      tensor_quant_params,
      op_quant_config,
      op_info,
      op_test_info,
      num_tensors,
      indices_to_ignore,
      is_inbounding_tensor,
      constant_inputs=None,
  ):
    """Tests all input or output tensors in provided quant params.

    Args:
      tensor_quant_params: Tensor transformation parameters.
      op_quant_config: Op quantization config.
      op_info: OpInfo object.
      op_test_info: OpTestInfo object.
      num_tensors: Number of tensors to test.
      indices_to_ignore: Indices of tensors to ignore.
      is_inbounding_tensor: Whether to test all inbounding tensors.
      constant_inputs: Constant inputs indices.
    """
    if not is_inbounding_tensor and constant_inputs is not None:
      raise ValueError(
          "Constant inputs should only be used for inbounding tensors."
      )
    constant_inputs = constant_inputs or []
    tensor_base_name = "input" if is_inbounding_tensor else "output"
    tensor_names = [tensor_base_name] + [
        f"{tensor_base_name}{i+2}" for i in range(num_tensors - 1)
    ]
    for i in range(num_tensors):
      if (
          # Check if SRQ.
          op_quant_config.compute_precision == _ComputePrecision.INTEGER
          and op_quant_config.activation_tensor_config is not None
          and i not in indices_to_ignore
      ):
        if is_inbounding_tensor:
          if i in constant_inputs:
            transformations = [_QuantTransformation.QUANTIZE_TENSOR]
          else:
            transformations = [_QuantTransformation.ADD_QUANTIZE]
        else:
          if i in constant_inputs:
            transformations = []
          else:
            transformations = [_QuantTransformation.ADD_DEQUANTIZE]
      else:
        transformations = [_QuantTransformation.NO_QUANTIZE]
      self._test_tensor_transformation_params(
          op_test_info.op_tensor_names[tensor_names[i]],
          op_info.subgraph_op_index,
          is_inbounding_tensor=is_inbounding_tensor,
          tensor_quant_config=op_quant_config.activation_tensor_config,
          transformation_params=tensor_quant_params[i],
          desired_transformations=transformations,
      )

  def _test_same_input_output_params(
      self,
      tensor_quant_params,
      num_inputs,
      num_outputs,
      inputs_to_ignore,
      outputs_to_ignore,
  ):
    """Tests input and output tensor transformation parameters are the same.

    Assumes that each of non-ignored input tensors has exactly one consumer.

    Args:
      tensor_quant_params: Tensor transformation parameters.
      num_inputs: Number of inputs in materialization function result.
      num_outputs: Number of outputs in materialization function result.
      inputs_to_ignore: Inputs to ignore.
      outputs_to_ignore: Outputs to ignore.
    """
    num_quant_inputs = num_inputs - len(inputs_to_ignore)
    num_quant_outputs = num_outputs - len(outputs_to_ignore)
    self.assertTrue(num_quant_inputs == 1 or num_quant_outputs == 1)

    # Test inputs. Ignoring `quantized_data` in comparison because inputs can be
    # constants and therefore have different quantized data.
    inputs_to_ignore = inputs_to_ignore or []
    input_quant_params = [
        dataclasses.replace(
            tensor_quant_params[i].consumers[0].parameters, quantized_data=None  # pytype: disable=attribute-error
        )
        for i in range(num_inputs)
        if i not in inputs_to_ignore
    ]
    expected_params = input_quant_params[0]
    for i in range(1, num_quant_inputs):
      self.assertEqual(input_quant_params[i], expected_params)

    # Test outputs.
    outputs_to_ignore = outputs_to_ignore or []
    for i in range(num_outputs):
      if i not in outputs_to_ignore:
        output_tensor_quant_params = tensor_quant_params[
            i + num_inputs
        ].producer.parameters  # pytype: disable=attribute-error
        self.assertEqual(output_tensor_quant_params, expected_params)

  def _test_no_weights_op(
      self,
      op_info,
      graph_info,
      op_test_info,
      materialization_func,
      get_tensor_quant_params_func,
      same_input_output_params=False,
      inputs_to_ignore=None,
      outputs_to_ignore=None,
      constant_inputs=None,
  ):
    """Test an op without weights and bias.

    Args:
      op_info: OpInfo object.
      graph_info: GraphInfo object.
      op_test_info: OpTestInfo object. In op_tensor_names field of this object,
        op tensor input names have to follow the pattern of "input", "input2",
        "input3" etc.
      materialization_func: Function to materialize tensor transformation
        parameters.
      get_tensor_quant_params_func: Function to get tensor quantization
        parameters.
      same_input_output_params: Whether the input and output tensor
        transformation parameters are the same.
      inputs_to_ignore: Inputs to ignore.
      outputs_to_ignore: Outputs to ignore.
      constant_inputs: Indices of constant inputs.
    """
    num_inputs = len(op_info.op.inputs)
    num_outputs = len(op_info.op.outputs)
    op_quant_config = op_info.op_quant_config
    self._setup_op_test_config(
        compute_precision=op_quant_config.compute_precision,
        is_drq=op_quant_config.activation_tensor_config is None,
        op_test_info=op_test_info,
        num_inputs=num_inputs,
        num_outputs=num_outputs,
    )
    tensor_quant_params = materialization_func(
        get_tensor_quant_params_func,
        op_info,
        graph_info,
        tensor_name_to_qsv=self._tensor_name_to_qsv,
        tensor_quant_params_cache=common_utils.TensorQuantParamsCache(),
    )
    self.assertLen(tensor_quant_params, num_inputs + num_outputs)

    # Test input tensor settings.
    inputs_to_ignore = inputs_to_ignore or []
    self._test_same_direction_tensors(
        tensor_quant_params[:num_inputs],
        op_quant_config,
        op_info,
        op_test_info,
        num_inputs,
        inputs_to_ignore,
        is_inbounding_tensor=True,
        constant_inputs=constant_inputs,
    )
    # Test output tensor settings.
    outputs_to_ignore = outputs_to_ignore or []
    self._test_same_direction_tensors(
        tensor_quant_params[num_inputs:],
        op_quant_config,
        op_info,
        op_test_info,
        num_outputs,
        outputs_to_ignore,
        is_inbounding_tensor=False,
    )
    # Test same input and output params.
    if same_input_output_params:
      self._test_same_input_output_params(
          tensor_quant_params,
          num_inputs,
          num_outputs,
          inputs_to_ignore,
          outputs_to_ignore,
      )
    # Test ignored inputs and outputs.
    self._test_ignored_inputs_and_outputs(
        tensor_quant_params,
        op_info,
        op_test_info,
        inputs_to_ignore,
        outputs_to_ignore,
    )

  def _test_fc_bmm_conv(
      self,
      op_info,
      graph_info,
      op_test_info,
      materialization_func,
      get_tensor_quant_params_func,
      bias_quantized_dim: int = 0,
      input_index: int = 0,
      weight_index: int = 1,
      bias_index: int = 2,
      inputs_to_ignore: Optional[Sequence[int]] = None,
      outputs_to_ignore: Optional[Sequence[int]] = None,
  ):
    """Tests fully connected, batch matmul and conv ops.

    Assumes that the op has one non-ignored output, which comes last in the
    materialization function result.

    Args:
      op_info: OpInfo object.
      graph_info: GraphInfo object.
      op_test_info: OpTestInfo object. In op_tensor_names field of this object,
        not ignored input tensors' names have to follow the pattern of "input",
        "input2", "input3" etc. Ignored input tensors' names have to follow the
        pattern of "ignored_input", "ignored_input2", "ignored_input3" etc.
      materialization_func: Function to materialize tensor transformation
        parameters.
      get_tensor_quant_params_func: Function to get tensor quantization
        parameters.
      bias_quantized_dim: Quantized dimension for bias.
      input_index: Input tensor index in the result of materialization function.
      weight_index: Weight tensor index in the result of materialization
        function.
      bias_index: Bias tensor index in the result of materialization function.
      inputs_to_ignore: Inputs to ignore.
      outputs_to_ignore: Outputs to ignore.
    """
    op_quant_config = op_info.op_quant_config
    self._setup_op_test_config(
        compute_precision=op_quant_config.compute_precision,
        is_drq=op_quant_config.activation_tensor_config is None,
        op_test_info=op_test_info,
    )
    tensor_quant_params = materialization_func(
        get_tensor_quant_params_func,
        op_info,
        graph_info,
        tensor_name_to_qsv=self._tensor_name_to_qsv,
        tensor_quant_params_cache=common_utils.TensorQuantParamsCache(),
    )

    _, weight_tensor, bias_tensor, _ = (
        tfl_flatbuffer_utils.parse_fc_bmm_conv_tensors(
            op_info.op,
            graph_info.subgraph_tensors,
        )
    )
    num_configs = 4 if bias_tensor is not None else 3
    if inputs_to_ignore:
      num_configs += len(inputs_to_ignore)
    if outputs_to_ignore:
      num_configs += len(outputs_to_ignore)
    self.assertLen(tensor_quant_params, num_configs)
    bias_tensor_data = None
    if bias_tensor is not None:
      bias_tensor_data = tfl_flatbuffer_utils.get_tensor_data(
          bias_tensor,
          op_test_info.test_model.buffers,
      )
    # Test input tensor settings
    transformations = [_QuantTransformation.NO_QUANTIZE]
    # Check if SRQ.
    if (
        op_quant_config.compute_precision == _ComputePrecision.INTEGER
        and op_quant_config.activation_tensor_config is not None
    ):
      transformations = [_QuantTransformation.ADD_QUANTIZE]
    self._test_tensor_transformation_params(
        op_test_info.op_tensor_names["input"],
        op_info.subgraph_op_index,
        is_inbounding_tensor=True,
        tensor_quant_config=op_quant_config.activation_tensor_config,
        transformation_params=tensor_quant_params[input_index],
        desired_transformations=transformations,
    )

    # Test weight tensor settings.
    transformations = [_QuantTransformation.QUANTIZE_TENSOR]
    # Check if WEIGHT_ONLY.
    if (
        op_quant_config.compute_precision == _ComputePrecision.FLOAT
        and op_quant_config.explicit_dequantize
    ):
      transformations = [
          _QuantTransformation.ADD_DEQUANTIZE,
      ]
    weight_tensor_data = tfl_flatbuffer_utils.get_tensor_data(
        weight_tensor,
        op_test_info.test_model.buffers,
    )
    self._test_tensor_transformation_params(
        op_test_info.op_tensor_names["weight"],
        op_info.subgraph_op_index,
        is_inbounding_tensor=True,
        tensor_quant_config=op_quant_config.weight_tensor_config,
        transformation_params=tensor_quant_params[weight_index],
        desired_transformations=transformations,
        tensor_data=weight_tensor_data,
        quantized_dimension=op_test_info.quantized_dimension,
    )

    # Test output tensor settings.
    transformations = [_QuantTransformation.NO_QUANTIZE]
    # Check if SRQ.
    if (
        op_quant_config.compute_precision == _ComputePrecision.INTEGER
        and op_quant_config.activation_tensor_config is not None
    ):
      transformations = [_QuantTransformation.ADD_DEQUANTIZE]
    self._test_tensor_transformation_params(
        op_test_info.op_tensor_names["output"],
        op_info.subgraph_op_index,
        is_inbounding_tensor=False,
        tensor_quant_config=op_quant_config.activation_tensor_config,
        transformation_params=tensor_quant_params[-1],
        desired_transformations=transformations,
    )

    # Test bias tensor settings.
    if bias_tensor is not None:
      bias_bit_width = 32
      transformations = [_QuantTransformation.NO_QUANTIZE]
      # Check if SRQ.
      if (
          op_quant_config.compute_precision == _ComputePrecision.INTEGER
          and op_quant_config.activation_tensor_config is not None
      ):
        transformations = [_QuantTransformation.QUANTIZE_TENSOR]
        if op_quant_config.activation_tensor_config.num_bits == 16:  # pytype: disable=attribute-error
          bias_bit_width = 64
      bias_config = qtyping.TensorQuantizationConfig(
          num_bits=bias_bit_width,
          symmetric=True,
          granularity=op_quant_config.weight_tensor_config.granularity,
      )
      self._test_tensor_transformation_params(
          op_test_info.op_tensor_names["bias"],
          op_info.subgraph_op_index,
          is_inbounding_tensor=True,
          tensor_quant_config=bias_config,
          transformation_params=tensor_quant_params[bias_index],
          desired_transformations=transformations,
          tensor_data=bias_tensor_data,
          quantized_dimension=bias_quantized_dim,
      )

    # Test ignored inputs and outputs.
    self._test_ignored_inputs_and_outputs(
        tensor_quant_params,
        op_info,
        op_test_info,
        inputs_to_ignore,
        outputs_to_ignore,
        is_op_with_weight=True,
    )

  def _get_ignore_tensor_name(
      self,
      ignored_tensor_index: int,
      is_op_with_weight: bool,
      is_inbounding_tensor: bool,
  ) -> str:
    """Gets the tensor name for ignored tensors.

    Args:
      ignored_tensor_index: Index of the ignored tensor in list of
        inputs/outputs.
      is_op_with_weight: Whether the op has weight.
      is_inbounding_tensor: Whether the tensor is an inbounding tensor.

    Returns:
      Tensor name.
    """
    base_name = "input" if is_inbounding_tensor else "output"
    if is_op_with_weight:
      base_name = f"ignored_{base_name}"
    if ignored_tensor_index == 0:
      return base_name
    return f"{base_name}{ignored_tensor_index+1}"

  def _test_ignored_inputs_and_outputs(
      self,
      tensor_quant_params,
      op_info,
      op_test_info,
      inputs_to_ignore: Optional[Sequence[int]] = None,
      outputs_to_ignore: Optional[Sequence[int]] = None,
      is_op_with_weight: bool = False,
  ):
    """Tests ignored inputs and outputs.

    Args:
      tensor_quant_params: Tensor transformation parameters.
      op_info: OpInfo object.
      op_test_info: OpTestInfo object.
      inputs_to_ignore: Inputs to ignore.
      outputs_to_ignore: Outputs to ignore.
      is_op_with_weight: Whether the op has weights.
    """
    op_quant_config = op_info.op_quant_config
    # Use activation tensor config just to pass something.
    tensor_quant_config = op_quant_config.activation_tensor_config
    inputs_to_ignore = inputs_to_ignore or []
    outputs_to_ignore = outputs_to_ignore or []
    # Test ignored inputs.
    for i in inputs_to_ignore:
      tensor_name = self._get_ignore_tensor_name(
          i, is_op_with_weight=is_op_with_weight, is_inbounding_tensor=True
      )
      self._test_tensor_transformation_params(
          op_test_info.op_tensor_names[tensor_name],
          op_info.subgraph_op_index,
          is_inbounding_tensor=True,
          tensor_quant_config=tensor_quant_config,
          transformation_params=tensor_quant_params[i],
          desired_transformations=[_QuantTransformation.NO_QUANTIZE],
      )
    # Test ignored outputs.
    for i in outputs_to_ignore:
      quant_params_index = len(op_info.op.inputs) + i
      tensor_name = self._get_ignore_tensor_name(
          i, is_op_with_weight=is_op_with_weight, is_inbounding_tensor=False
      )
      self._test_tensor_transformation_params(
          op_test_info.op_tensor_names[tensor_name],
          op_info.subgraph_op_index,
          is_inbounding_tensor=False,
          tensor_quant_config=tensor_quant_config,
          transformation_params=tensor_quant_params[quant_params_index],
          desired_transformations=[_QuantTransformation.NO_QUANTIZE],
      )

  def _test_tensor_transformation_params(
      self,
      tensor_name,
      subgraph_op_id,
      is_inbounding_tensor,
      tensor_quant_config,
      transformation_params,
      desired_transformations,
      tensor_data=None,
      expected_tensor_max=None,
      quantized_dimension=0,
  ):
    """Tests tensor transformation parameters.

    Args:
      tensor_name: Tensor name.
      subgraph_op_id: Subgraph op id.
      is_inbounding_tensor: Whether the tensor is an inbounding tensor.
      tensor_quant_config: Tensor quantization config.
      transformation_params: Tensor transformation parameters.
      desired_transformations: Desired transformations.
      tensor_data: Tensor data.
      expected_tensor_max: Expected tensor max.
      quantized_dimension: Quantized dimension.
    """
    self.assertEqual(transformation_params.tensor_name, tensor_name)
    if is_inbounding_tensor:
      self.assertIsNone(transformation_params.producer)
      self.assertLen(transformation_params.consumers, 1)
      op_params = transformation_params.consumers[0]
    else:
      self.assertIsNone(transformation_params.consumers)
      op_params = transformation_params.producer
      self.assertIsNotNone(op_params)
    self.assertEqual(op_params.subgraph_op_id, subgraph_op_id)
    self.assertSequenceEqual(op_params.transformations, desired_transformations)
    quantization_params = op_params.parameters
    if desired_transformations == [_QuantTransformation.NO_QUANTIZE]:
      self.assertIsNone(quantization_params)
    else:
      self.assertIsNotNone(quantization_params)
      if (
          tensor_quant_config.granularity
          is qtyping.QuantGranularity.CHANNELWISE
      ):
        self.assertEqual(
            quantization_params.quantized_dimension, quantized_dimension
        )
      else:
        self.assertIsNone(quantization_params.quantized_dimension)
      self.assertEqual(
          quantization_params.num_bits, tensor_quant_config.num_bits
      )
      # Test correctness of quantization parameters.
      if tensor_data is not None:
        expected_quantized_data = uniform_quantize_tensor.uniform_quantize(
            tensor_data, quantization_params
        )
        self.assertSequenceEqual(
            list(expected_quantized_data.flatten()),
            list(quantization_params.quantized_data.flatten()),  # pytype: disable=attribute-error
        )
      elif expected_tensor_max:
        max_q = 2**tensor_quant_config.num_bits / 2 - 1
        calculated_tensor_max = quantization_params.scale[0] * (
            max_q - quantization_params.zero_point[0]
        )
        self.assertAlmostEqual(
            calculated_tensor_max, expected_tensor_max, delta=5e-2
        )

  def _test_materialize_fn_quantizes_weights_larger_than_min_weight_elements_for_w8_afp32(
      self,
      op_name: qtyping.TFLOperationName,
      subgraph_op_id: int,
      min_weight_elements: int,
      graph_info: qtyping.GraphInfo,
      op_test_info: OpTestInfo,
      materialization_func: Callable[..., Any],
      get_tensor_quant_params_func: Callable[..., Any],
      expect_weights_quantized: bool,
  ):
    """Tests weights are quantized if they are larger than min_weight_elements.

    Args:
      op_name: Op name.
      subgraph_op_id: Subgraph op id.
      min_weight_elements: Minimum number of weight elements.
      graph_info: GraphInfo object.
      op_test_info: OpTestInfo object.
      materialization_func: Function to materialize tensor transformation
        parameters.
      get_tensor_quant_params_func: Function to get tensor quantization
        parameters.
      expect_weights_quantized: Whether the weights are expected to be
        quantized.
    """
    subgraph0 = op_test_info.test_model.subgraphs[0]
    op = subgraph0.operators[subgraph_op_id]
    op_info = qtyping.OpInfo(
        op=op,
        op_name=op_name,
        subgraph_op_index=subgraph_op_id,
        op_quant_config=qtyping.OpQuantizationConfig(
            activation_tensor_config=None,
            weight_tensor_config=_TensorQuantConfig(
                num_bits=8,
                symmetric=True,
                granularity=qtyping.QuantGranularity.CHANNELWISE,
            ),
            compute_precision=_ComputePrecision.INTEGER,
            explicit_dequantize=False,
            skip_checks=False,
            min_weight_elements=min_weight_elements,
        ),
    )
    _, weight_quant_params, *_ = materialization_func(
        get_tensor_quant_params_func,
        op_info,
        graph_info,
        tensor_name_to_qsv=self._tensor_name_to_qsv,
        tensor_quant_params_cache=common_utils.TensorQuantParamsCache(),
    )
    self.assertEqual(
        weight_quant_params.tensor_name, op_test_info.op_tensor_names["weight"]
    )
    self.assertLen(weight_quant_params.consumers, 1)
    quantization_params = weight_quant_params.consumers[0].parameters
    transformations = weight_quant_params.consumers[0].transformations
    if expect_weights_quantized:
      self.assertIsNotNone(quantization_params)
      self.assertEqual(
          transformations,
          [_QuantTransformation.QUANTIZE_TENSOR],
      )
    else:
      self.assertIsNone(quantization_params)
      self.assertEqual(
          transformations,
          [_QuantTransformation.NO_QUANTIZE],
      )



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import pathlib
from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.nonlinear_quantize import float_casting
from ai_edge_quantizer.algorithms.utils import common_utils
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils

_TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile("../../tests/models")
_TFLOpName = qtyping.TFLOperationName
_ComputePrecision = qtyping.ComputePrecision
_TensorQuantConfig = qtyping.TensorQuantizationConfig
_QuantTransformation = qtyping.QuantTransformation


class Fp16QuantizeTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    np.random.seed(666)
    self._test_model_path = str(
        pathlib.Path(_TEST_DATA_PREFIX_PATH) / "conv_fc_mnist.tflite"
    )
    self._test_model = tfl_flatbuffer_utils.read_model(self._test_model_path)
    # The test model has one subgraph for now.
    self._graph_info = qtyping.GraphInfo(
        subgraph_tensors=self._test_model.subgraphs[0].tensors,
        buffers=self._test_model.buffers,
    )
    self._tensor_name_to_qsv = {}

  @parameterized.named_parameters(
      dict(
          testcase_name="fc",
          op_name=_TFLOpName.FULLY_CONNECTED,
      ),
      dict(
          testcase_name="conv2d",
          op_name=_TFLOpName.CONV_2D,
      ),
      dict(
          testcase_name="embedding_lookup",
          op_name=_TFLOpName.EMBEDDING_LOOKUP,
      ),
  )
  def test_check_op_quantization_config_succeeds(self, op_name):
    float_casting.check_op_quantization_config(
        op_name,
        qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(
                num_bits=16, dtype=qtyping.TensorDataType.FLOAT
            ),
            compute_precision=_ComputePrecision.FLOAT,  # WEIGHT_ONLY.
            explicit_dequantize=True,
        ),
    )

  @parameterized.named_parameters(
      dict(
          testcase_name="invalid_fc",
          op_name=_TFLOpName.FULLY_CONNECTED,
      ),
      dict(
          testcase_name="invalid_conv2d",
          op_name=_TFLOpName.CONV_2D,
      ),
      dict(
          testcase_name="invalid_embedding_lookup",
          op_name=_TFLOpName.EMBEDDING_LOOKUP,
      ),
  )
  def test_check_op_quantization_config_invalid_activation_tensor_config_raises_exception(
      self, op_name
  ):
    # With activation tensor config.
    error_message = (
        "Activation tensor quantization is not supported for float casting"
        " quantization."
    )
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      float_casting.check_op_quantization_config(
          op_name,
          qtyping.OpQuantizationConfig(
              activation_tensor_config=_TensorQuantConfig(
                  num_bits=16, dtype=qtyping.TensorDataType.FLOAT
              ),
              weight_tensor_config=_TensorQuantConfig(
                  num_bits=16, dtype=qtyping.TensorDataType.FLOAT
              ),
              compute_precision=_ComputePrecision.FLOAT,  # WEIGHT_ONLY.
              explicit_dequantize=True,
          ),
      )

  @parameterized.named_parameters(
      dict(
          testcase_name="invalid_fc",
          op_name=_TFLOpName.FULLY_CONNECTED,
      ),
      dict(
          testcase_name="invalid_conv2d",
          op_name=_TFLOpName.CONV_2D,
      ),
      dict(
          testcase_name="invalid_embedding_lookup",
          op_name=_TFLOpName.EMBEDDING_LOOKUP,
      ),
  )
  def test_check_op_quantization_config_invalid_bit_width_raises_exception(
      self, op_name
  ):
    error_message = (
        "float casting quantization config requires number of bits to be set as"
        " 16, dtype as float"
    )
    # Wrong bit width.
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      float_casting.check_op_quantization_config(
          op_name,
          qtyping.OpQuantizationConfig(
              activation_tensor_config=None,
              weight_tensor_config=_TensorQuantConfig(
                  num_bits=8, dtype=qtyping.TensorDataType.FLOAT
              ),
              compute_precision=_ComputePrecision.FLOAT,  # WEIGHT_ONLY.
              explicit_dequantize=True,
          ),
      )

  @parameterized.named_parameters(
      dict(
          testcase_name="invalid_fc",
          op_name=_TFLOpName.FULLY_CONNECTED,
      ),
      dict(
          testcase_name="invalid_conv2d",
          op_name=_TFLOpName.CONV_2D,
      ),
      dict(
          testcase_name="invalid_embedding_lookup",
          op_name=_TFLOpName.EMBEDDING_LOOKUP,
      ),
  )
  def test_check_op_quantization_config_invalid_dtype_raises_exception(
      self, op_name
  ):
    error_message = (
        "float casting quantization config requires number of bits to be set as"
        " 16, dtype as float"
    )
    # Wrong dtype.
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      float_casting.check_op_quantization_config(
          op_name,
          qtyping.OpQuantizationConfig(
              activation_tensor_config=None,
              weight_tensor_config=_TensorQuantConfig(
                  num_bits=16, dtype=qtyping.TensorDataType.INT
              ),
              compute_precision=_ComputePrecision.FLOAT,  # WEIGHT_ONLY.
              explicit_dequantize=True,
          ),
      )

  @parameterized.named_parameters(
      dict(
          testcase_name="invalid_fc",
          op_name=_TFLOpName.FULLY_CONNECTED,
      ),
      dict(
          testcase_name="invalid_conv2d",
          op_name=_TFLOpName.CONV_2D,
      ),
      dict(
          testcase_name="invalid_embedding_lookup",
          op_name=_TFLOpName.EMBEDDING_LOOKUP,
      ),
  )
  def test_check_op_quantization_config_no_weight_config_raises_exception(
      self, op_name
  ):
    error_message = (
        "Weight tensor quantization config is required for float casting"
        " quantization."
    )
    # No weight quantization config.
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      float_casting.check_op_quantization_config(
          op_name,
          qtyping.OpQuantizationConfig(
              activation_tensor_config=None,
              compute_precision=_ComputePrecision.FLOAT,  # WEIGHT_ONLY.
              explicit_dequantize=True,
          ),
      )

  @parameterized.named_parameters(
      dict(
          testcase_name="averagepool2D",
          op_name=_TFLOpName.AVERAGE_POOL_2D,
      ),
      dict(
          testcase_name="reshape",
          op_name=_TFLOpName.RESHAPE,
      ),
  )
  def test_check_op_quantization_config_invalid_ops_raises_exception(
      self, op_name
  ):
    error_message = "Unsupported op"
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      float_casting.check_op_quantization_config(
          op_name=op_name,
          op_quant_config=qtyping.OpQuantizationConfig(
              activation_tensor_config=None,
              weight_tensor_config=_TensorQuantConfig(
                  num_bits=16, dtype=qtyping.TensorDataType.FLOAT
              ),
              compute_precision=_ComputePrecision.FLOAT,  # WEIGHT_ONLY.
              explicit_dequantize=True,
          ),
      )

  @parameterized.named_parameters(
      dict(
          testcase_name="fc_with_bias",
          subgraph_op_id=3,
          op_tensor_names={
              "weight": "arith.constant1",
              "bias": "arith.constant2",
              "input": "sequential/flatten/Reshape",
              "output": (
                  "sequential/dense/MatMul;sequential/dense/Relu;sequential/dense/BiasAdd"
              ),
          },
      ),
      dict(
          testcase_name="fc_with_no_bias",
          subgraph_op_id=4,
          op_tensor_names={
              "weight": "arith.constant",
              "input": (
                  "sequential/dense/MatMul;sequential/dense/Relu;sequential/dense/BiasAdd"
              ),
              "output": "sequential/dense_1/MatMul",
          },
      ),
  )
  def test_fully_connected_weight_only_succeeds(
      self, subgraph_op_id, op_tensor_names
  ):
    subgraph0 = self._test_model.subgraphs[0]
    fc_op = subgraph0.operators[subgraph_op_id]
    op_info = qtyping.OpInfo(
        op=fc_op,
        op_name=_TFLOpName.FULLY_CONNECTED,
        subgraph_op_index=subgraph_op_id,
        op_quant_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(
                num_bits=16, dtype=qtyping.TensorDataType.FLOAT
            ),
            compute_precision=_ComputePrecision.FLOAT,  # WEIGHT_ONLY.
            explicit_dequantize=True,
        ),
    )

    self._test_fc_conv(
        op_info,
        self._graph_info,
        op_tensor_names,
        float_casting.materialize_fc_conv,
    )

  def test_conv2d_weight_only_succeeds(self):
    # Read from Model Explorer.
    subgraph0 = self._test_model.subgraphs[0]
    subgraph_op_id = 0
    op = subgraph0.operators[subgraph_op_id]

    op_info = qtyping.OpInfo(
        op=op,
        op_name=_TFLOpName.CONV_2D,
        subgraph_op_index=subgraph_op_id,
        op_quant_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(
                num_bits=16, dtype=qtyping.TensorDataType.FLOAT
            ),
            compute_precision=_ComputePrecision.FLOAT,  # WEIGHT_ONLY.
            explicit_dequantize=True,
        ),
    )

    op_tensor_names = {}
    op_tensor_names["weight"] = "sequential/conv2d/Conv2D"
    op_tensor_names["bias"] = (
        "sequential/conv2d/Relu;sequential/conv2d/BiasAdd;sequential/conv2d/Conv2D;sequential/conv2d/BiasAdd/ReadVariableOp"
    )
    op_tensor_names["input"] = "serving_default_conv2d_input:0"
    op_tensor_names["output"] = (
        "sequential/conv2d/Relu;sequential/conv2d/BiasAdd;sequential/conv2d/Conv2D;sequential/conv2d/BiasAdd/ReadVariableOp1"
    )
    self._test_fc_conv(
        op_info,
        self._graph_info,
        op_tensor_names,
        float_casting.materialize_fc_conv,
    )

  def test_conv2d_dynamic_weight_succeeds(self):
    # Read from Model Explorer.
    subgraph0 = self._test_model.subgraphs[0]
    subgraph_op_id = 0
    op = subgraph0.operators[subgraph_op_id]

    op_info = qtyping.OpInfo(
        op=op,
        op_name=_TFLOpName.CONV_2D,
        subgraph_op_index=subgraph_op_id,
        op_quant_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(
                num_bits=16, dtype=qtyping.TensorDataType.FLOAT
            ),
            compute_precision=_ComputePrecision.FLOAT,  # WEIGHT_ONLY.
            explicit_dequantize=True,
        ),
    )

    with mock.patch.object(
        tfl_flatbuffer_utils, "get_tensor_data", return_value=None
    ):
      tensor_quant_params = float_casting.materialize_fc_conv(
          op_info,
          self._graph_info,
          self._tensor_name_to_qsv,
          tensor_quant_params_cache=common_utils.TensorQuantParamsCache(),
      )

    # 4 tensors (input, weight, bias, output)
    self.assertLen(tensor_quant_params, 4)
    # The weight tensor is at index 1.
    weight_params = tensor_quant_params[1]
    self.assertEqual(weight_params.tensor_name, "sequential/conv2d/Conv2D")
    self.assertLen(weight_params.consumers, 1)

    # Transformation should be NO_QUANTIZE since we mocked get_tensor_data to
    # None.
    op_params = weight_params.consumers[0]
    self.assertSequenceEqual(
        op_params.transformations, [_QuantTransformation.NO_QUANTIZE]
    )
    self.assertIsNone(op_params.parameters)

  @parameterized.named_parameters(
      dict(
          testcase_name="invalid_fc",
          op_name=_TFLOpName.FULLY_CONNECTED,
      ),
      dict(
          testcase_name="invalid_conv2d",
          op_name=_TFLOpName.CONV_2D,
      ),
      dict(
          testcase_name="invalid_embedding_lookup",
          op_name=_TFLOpName.EMBEDDING_LOOKUP,
      ),
  )
  def test_check_op_quantization_config_invalid_execution_mode_raises_exception(
      self, op_name
  ):
    # Use DRQ instead of WEIGHT-ONLY.
    error_message = (
        "Currently, only Weight-Only is supported for float casting"
        " quantization."
    )
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      float_casting.check_op_quantization_config(
          op_name,
          qtyping.OpQuantizationConfig(
              activation_tensor_config=None,
              weight_tensor_config=_TensorQuantConfig(
                  num_bits=16, dtype=qtyping.TensorDataType.FLOAT
              ),
              compute_precision=_ComputePrecision.INTEGER,  # DRQ.
          ),
      )

  def test_conv2d_transpose_weight_only_succeeds(self):
    # Read from Model Explorer.
    test_model_path = str(
        pathlib.Path(_TEST_DATA_PREFIX_PATH)
        / "single_conv2d_transpose_bias.tflite"
    )

    test_model = tfl_flatbuffer_utils.read_model(test_model_path)
    # The test model has one subgraph for now.
    graph_info = qtyping.GraphInfo(
        subgraph_tensors=test_model.subgraphs[0].tensors,
        buffers=test_model.buffers,
    )

    subgraph0 = test_model.subgraphs[0]
    subgraph_op_id = 0
    op = subgraph0.operators[subgraph_op_id]

    op_info = qtyping.OpInfo(
        op=op,
        op_name=_TFLOpName.CONV_2D_TRANSPOSE,
        subgraph_op_index=subgraph_op_id,
        op_quant_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(
                num_bits=16, dtype=qtyping.TensorDataType.FLOAT
            ),
            compute_precision=_ComputePrecision.FLOAT,  # WEIGHT_ONLY.
            explicit_dequantize=True,
        ),
    )

    op_tensor_names = {}
    op_tensor_names["weight"] = (
        "sequential_5/conv2d_transpose_3/conv2d_transpose"
    )
    op_tensor_names["bias"] = (
        "sequential_5/conv2d_transpose_3/BiasAdd;sequential_5/conv2d_transpose_3/conv2d_transpose;sequential_5/conv2d_transpose_3/BiasAdd/ReadVariableOp"
    )
    op_tensor_names["input"] = "serving_default_input_6:0"
    op_tensor_names["output"] = "StatefulPartitionedCall:0"

    tensor_quant_params = float_casting.materialize_conv2d_transpose(
        op_info,
        graph_info,
        self._tensor_name_to_qsv,
        tensor_quant_params_cache=common_utils.TensorQuantParamsCache(),
    )
    _, weight_tensor, bias_tensor, _ = (
        tfl_flatbuffer_utils.parse_fc_bmm_conv_tensors(
            op_info.op, graph_info.subgraph_tensors
        )
    )

    num_configs = 4 if bias_tensor is not None else 3
    self.assertLen(tensor_quant_params, num_configs)

    # Test input tensor params.
    self._test_fp16_nonweight_tensor_transformation_params(
        op_tensor_names["input"],
        op_info.subgraph_op_index,
        transformation_params=tensor_quant_params[0],
        desired_transformations=[_QuantTransformation.NO_QUANTIZE],
        is_inbounding_tensor=True,
    )

    # Test weight tensor params.
    weight_tensor_data = tfl_flatbuffer_utils.get_tensor_data(
        weight_tensor,
        graph_info.buffers,
    )
    self._test_fp16_weight_tensor_transformation_params(
        op_tensor_names["weight"],
        op_info.subgraph_op_index,
        tensor_quant_config=op_info.op_quant_config.weight_tensor_config,
        transformation_params=tensor_quant_params[1],
        desired_transformations=[_QuantTransformation.ADD_DEQUANTIZE],
        tensor_data=weight_tensor_data,
    )
    # Test output tensor params.
    self._test_fp16_nonweight_tensor_transformation_params(
        op_tensor_names["output"],
        op_info.subgraph_op_index,
        transformation_params=tensor_quant_params[2],
        desired_transformations=[_QuantTransformation.NO_QUANTIZE],
        is_inbounding_tensor=False,
    )

    # Test bias tensor params.
    if bias_tensor is not None:
      self._test_fp16_nonweight_tensor_transformation_params(
          op_tensor_names["bias"],
          op_info.subgraph_op_index,
          transformation_params=tensor_quant_params[3],
          desired_transformations=[_QuantTransformation.NO_QUANTIZE],
          is_inbounding_tensor=True,
      )

  def test_conv2d_transpose_dynamic_weight_succeeds(self):
    # Read from Model Explorer.
    test_model_path = str(
        pathlib.Path(_TEST_DATA_PREFIX_PATH)
        / "single_conv2d_transpose_bias.tflite"
    )

    test_model = tfl_flatbuffer_utils.read_model(test_model_path)
    # The test model has one subgraph for now.
    graph_info = qtyping.GraphInfo(
        subgraph_tensors=test_model.subgraphs[0].tensors,
        buffers=test_model.buffers,
    )

    subgraph0 = test_model.subgraphs[0]
    subgraph_op_id = 0
    op = subgraph0.operators[subgraph_op_id]

    op_info = qtyping.OpInfo(
        op=op,
        op_name=_TFLOpName.CONV_2D_TRANSPOSE,
        subgraph_op_index=subgraph_op_id,
        op_quant_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(
                num_bits=16, dtype=qtyping.TensorDataType.FLOAT
            ),
            compute_precision=_ComputePrecision.FLOAT,  # WEIGHT_ONLY.
            explicit_dequantize=True,
        ),
    )

    with mock.patch.object(
        tfl_flatbuffer_utils, "get_tensor_data", return_value=None
    ):
      tensor_quant_params = float_casting.materialize_conv2d_transpose(
          op_info,
          graph_info,
          self._tensor_name_to_qsv,
          tensor_quant_params_cache=common_utils.TensorQuantParamsCache(),
      )

    _, _, bias_tensor, _ = tfl_flatbuffer_utils.parse_fc_bmm_conv_tensors(
        op_info.op, graph_info.subgraph_tensors
    )

    num_configs = 4 if bias_tensor is not None else 3
    self.assertLen(tensor_quant_params, num_configs)

    # The weight tensor is at index 1.
    weight_params = tensor_quant_params[1]
    self.assertEqual(
        weight_params.tensor_name,
        "sequential_5/conv2d_transpose_3/conv2d_transpose",
    )
    self.assertLen(weight_params.consumers, 1)

    # Transformation should be NO_QUANTIZE since we mocked get_tensor_data to
    # None.
    op_params = weight_params.consumers[0]
    self.assertSequenceEqual(
        op_params.transformations, [_QuantTransformation.NO_QUANTIZE]
    )
    self.assertIsNone(op_params.parameters)

  def test_depthwise_conv2d_weight_only_succeeds(self):
    # Read from Model Explorer.
    test_model_path = str(
        pathlib.Path(_TEST_DATA_PREFIX_PATH)
        / "single_depthwise_conv2d_bias.tflite"
    )

    test_model = tfl_flatbuffer_utils.read_model(test_model_path)
    # The test model has one subgraph for now.
    graph_info = qtyping.GraphInfo(
        subgraph_tensors=test_model.subgraphs[0].tensors,
        buffers=test_model.buffers,
    )

    subgraph0 = test_model.subgraphs[0]
    subgraph_op_id = 0
    op = subgraph0.operators[subgraph_op_id]

    op_info = qtyping.OpInfo(
        op=op,
        op_name=_TFLOpName.DEPTHWISE_CONV_2D,
        subgraph_op_index=subgraph_op_id,
        op_quant_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(
                num_bits=16, dtype=qtyping.TensorDataType.FLOAT
            ),
            compute_precision=_ComputePrecision.FLOAT,  # WEIGHT_ONLY.
            explicit_dequantize=True,
        ),
    )

    op_tensor_names = {}
    op_tensor_names["weight"] = "sequential/depthwise_conv2d/depthwise"
    op_tensor_names["bias"] = (
        "sequential/depthwise_conv2d/BiasAdd;sequential/depthwise_conv2d/depthwise;sequential/depthwise_conv2d/BiasAdd/ReadVariableOp"
    )
    op_tensor_names["input"] = "serving_default_input_1:0"
    op_tensor_names["output"] = "StatefulPartitionedCall:0"
    self._test_fc_conv(
        op_info,
        graph_info,
        op_tensor_names,
        float_casting.materialize_fc_conv,
    )

  def test_embedding_lookup_weight_only_succeeds(self):
    test_model_path = str(
        pathlib.Path(_TEST_DATA_PREFIX_PATH) / "embedding_lookup.tflite"
    )

    test_model = tfl_flatbuffer_utils.read_model(test_model_path)
    graph_info = qtyping.GraphInfo(
        subgraph_tensors=test_model.subgraphs[0].tensors,
        buffers=test_model.buffers,
    )

    subgraph0 = test_model.subgraphs[0]
    subgraph_op_id = 0
    op = subgraph0.operators[subgraph_op_id]

    op_info = qtyping.OpInfo(
        op=op,
        op_name=_TFLOpName.EMBEDDING_LOOKUP,
        subgraph_op_index=subgraph_op_id,
        op_quant_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(
                num_bits=16, dtype=qtyping.TensorDataType.FLOAT
            ),
            compute_precision=_ComputePrecision.FLOAT,  # WEIGHT_ONLY.
            explicit_dequantize=True,
        ),
    )

    op_tensor_names = {}
    op_tensor_names["weight"] = (
        "jit(export_func)/jit(main)/...y,yz->...z/dot_general;jit(export_func)/jit(main)/jit(_one_hot)/eq;jit(export_func)/jit(main)/jit(_one_hot)/convert_element_type"
    )
    op_tensor_names["input"] = "lookup"
    op_tensor_names["output"] = "Identity_1"

    # TODO: b/335913710 - Rename the test function.
    self._test_fc_conv(
        op_info,
        graph_info,
        op_tensor_names,
        float_casting.materialize_fc_conv,
    )

  def _test_fc_conv(
      self,
      op_info,
      graph_info,
      op_tensor_names,
      materialization_func,
  ):

    tensor_quant_params = materialization_func(
        op_info,
        graph_info,
        self._tensor_name_to_qsv,
        tensor_quant_params_cache=common_utils.TensorQuantParamsCache(),
    )
    _, weight_tensor, bias_tensor, _ = (
        tfl_flatbuffer_utils.parse_fc_bmm_conv_tensors(
            op_info.op, graph_info.subgraph_tensors
        )
    )

    num_configs = 4 if bias_tensor is not None else 3
    self.assertLen(tensor_quant_params, num_configs)

    # Test input tensor params.
    self._test_fp16_nonweight_tensor_transformation_params(
        op_tensor_names["input"],
        op_info.subgraph_op_index,
        transformation_params=tensor_quant_params[0],
        desired_transformations=[_QuantTransformation.NO_QUANTIZE],
        is_inbounding_tensor=True,
    )

    # Test weight tensor params.
    weight_tensor_data = tfl_flatbuffer_utils.get_tensor_data(
        weight_tensor,
        graph_info.buffers,
    )

    self._test_fp16_weight_tensor_transformation_params(
        op_tensor_names["weight"],
        op_info.subgraph_op_index,
        tensor_quant_config=op_info.op_quant_config.weight_tensor_config,
        transformation_params=tensor_quant_params[1],
        desired_transformations=[_QuantTransformation.ADD_DEQUANTIZE],
        tensor_data=weight_tensor_data,
    )
    # Test output tensor params.
    self._test_fp16_nonweight_tensor_transformation_params(
        op_tensor_names["output"],
        op_info.subgraph_op_index,
        transformation_params=tensor_quant_params[2],
        desired_transformations=[_QuantTransformation.NO_QUANTIZE],
        is_inbounding_tensor=False,
    )

    # Test bias tensor params.
    if bias_tensor is not None:
      self._test_fp16_nonweight_tensor_transformation_params(
          op_tensor_names["bias"],
          op_info.subgraph_op_index,
          transformation_params=tensor_quant_params[3],
          desired_transformations=[_QuantTransformation.NO_QUANTIZE],
          is_inbounding_tensor=True,
      )

  def _test_fp16_weight_tensor_transformation_params(
      self,
      tensor_name,
      subgraph_op_id,
      tensor_quant_config,
      transformation_params,
      desired_transformations,
      tensor_data,
  ):
    self.assertEqual(transformation_params.tensor_name, tensor_name)
    # Weight-only means the transformation is added from the consumer.
    self.assertIsNone(transformation_params.producer)
    self.assertLen(transformation_params.consumers, 1)
    # Check op params.
    op_params = transformation_params.consumers[0]
    self.assertEqual(op_params.subgraph_op_id, subgraph_op_id)
    self.assertSequenceEqual(op_params.transformations, desired_transformations)
    # Check quantization params.
    quantization_params = op_params.parameters
    self.assertIsNotNone(quantization_params)
    self.assertIsNotNone(tensor_quant_config)
    self.assertEqual(quantization_params.num_bits, tensor_quant_config.num_bits)
    quantized_data = quantization_params.quantized_data
    self.assertIsNotNone(quantized_data)
    self.assertEqual(quantized_data.dtype, "float16")
    # fp16 quantization implies very small error.
    self.assertSequenceAlmostEqual(
        list(tensor_data.flatten()),  # pytype: disable=attribute-error
        list(quantization_params.quantized_data.flatten()),  # pytype: disable=attribute-error
        delta=5,
    )

  def _test_fp16_nonweight_tensor_transformation_params(
      self,
      tensor_name,
      subgraph_op_id,
      transformation_params,
      desired_transformations,
      is_inbounding_tensor,
  ):
    self.assertEqual(transformation_params.tensor_name, tensor_name)
    if is_inbounding_tensor:
      self.assertIsNone(transformation_params.producer)
      self.assertLen(transformation_params.consumers, 1)
      op_params = transformation_params.consumers[0]
    else:
      self.assertIsNone(transformation_params.consumers)
      op_params = transformation_params.producer
    self.assertIsNotNone(op_params)
    self.assertEqual(op_params.subgraph_op_id, subgraph_op_id)
    self.assertSequenceEqual(op_params.transformations, desired_transformations)
    self.assertIsNone(op_params.parameters)


if __name__ == "__main__":
  absltest.main()
